import tkinter as tk

class ServiceExit(Exception):
    """
    Custom exception which is used to trigger the clean exit
    of all running threads and the main program.
    """
    pass

def service_shutdown(signum, frame):
    raise ServiceExit

class Window(tk.Tk):
    def __init__(self,rows=1,cols=1,*args,**kwargs):
        super().__init__(*args,**kwargs)

    def stop(self):
        self.destroy()
    def start(self):
        self.mainloop()



from pysts.utils.modules import Lazy_Module
from .listener import Listener
from pysts.utils.utils import append_fcns

pynput=Lazy_Module('pynput')

class InputListener(Listener):
    _callbacks=None
    def get_callbacks(self,*args,**kwargs):
        if not self._callbacks:
            self._callbacks={}
        return self._callbacks
    def add_func_to_dict(self,obj,key,func):
        if key in obj:
            func=append_fcns(obj[key])
        obj[key]=func

class Keyboard(InputListener):
    _hotkeys=None
    def get_hotkeys(self,*args,**kwargs):
        if not self._hotkeys:
            self._hotkeys={}
        return self._hotkeys
    def create_hotkey_listener(self):
        return pynput.keyboard.GlobalHotKeys(self.get_hotkeys())
    def create_listener(self):
        return pynput.keyboard.Listener(**self.get_callbacks())
    @Listener.event
    def on_hotkey(self,func,hotkey,*args,**kwargs):
        self.add_func_to_dict(self.get_hotkeys(),hotkey,func)
        return self.create_hotkey_listener
    @Listener.event
    def on_press(self,func,*args,**kwargs): #key
        self.add_func_to_dict(self.get_callbacks(),'on_press',func)
        return self.create_listener
    @Listener.event
    def on_release(self,func,*args,**kwargs): #key
        self.add_func_to_dict(self.get_callbacks(),'on_release',func)
        return self.create_listener

class Mouse(InputListener):
    def create_listener(self):
        return pynput.mouse.Listener(**self.get_callbacks())
    @Listener.event
    def on_move(self,func,*args,**kwargs): #x, y
        self.add_func_to_dict(self.get_callbacks(),'on_move',func)
        return self.create_listener
    @Listener.event
    def on_click(self,func,*args,**kwargs): #x, y, button, pressed
        self.add_func_to_dict(self.get_callbacks(),'on_click',func)
        return self.create_listener
    @Listener.event
    def on_scroll(self,func,*args,**kwargs): #x, y, dx, dy
        self.add_func_to_dict(self.get_callbacks(),'on_scroll',func)
        return self.create_listener

if __name__ == "__main__":
    keylistener = Keyboard()
    @keylistener.on_press
    def print_key_presses(key):
        vktext=f' with vk {key.vk}' if hasattr(key,'vk') else ''
        try:
            print('alphanumeric key {0} pressed{1}'.format(
                key.char,vktext))
        except AttributeError:
            print('special key {0} pressed{1}'.format(
                key,vktext))

    @keylistener.on_hotkey('<ctrl>+<alt>+q')
    def print_hotkey():
        print('HOTKEY: <ctrl>+<alt>+q pressed')


    mouselistener = Mouse()
    @mouselistener.on_click
    def print_click(x, y, button, pressed):
        print(f"Mouse button {button} {'pressed' if pressed else 'released'} at ({x}, {y})")

    Listener._start_jobs(block=True)



from .. import components
from ..components.base import *
from ..config import * #METRICS_PATH
import pandas as pd
import numpy as np

class Overview(components.Component):
    def __init__(self):

        self.metrics=pd.read_csv(METRICS_PATH).astype({'Date':'datetime64'})

        units=self.metrics.SN.to_list()
        unit_options=[{'label': x, 'value': x} for x in units]
        display_options=[{'label': 'All Units', 'value': 'all'}]+unit_options
        highlight_options=[{'label': 'Latest Unit', 'value': 'latest'},{'label': 'None', 'value': 'none'}]+unit_options
        self.display=components.Dropdown(multi=True,label='Display',options=display_options,update_value_on_change=True,keep_old_values=True,value='all')
        self.highlight=components.Dropdown(multi=False,label='Highlight',options=highlight_options,update_value_on_change=True,keep_old_values=True,value='latest')

        def get_figs(highlights,metrics=None):
            if metrics is None: metrics=self.metrics
            if highlights is None: highlights=[]

            figs=[None]*4
            fig_cols=[('Coupon','Particle'),('Coupon SBR','Particle SBR'),('Coupon','Coupon Background'),('Particle','Particle Background')]
            renames={'Coupon':'Coupon 315.5nm Fe Intensity (au)',
                    'Coupon Background':'Coupon 315.5nm Fe Background',
                    'Coupon SBR':'Coupon 315.5nm Fe SBR',
                    'Particle':'Particle 315.5nm Fe Intensity (au)',
                    'Particle Background':'Particle 315.5nm Fe Background',
                    'Particle SBR':'Particle 315.5nm Fe SBR'}
            #metrics=metrics.rename(columns=renames)
            figs=[go.Figure(layout={'showlegend':False,'height':500,'xaxis_title':renames[fig_cols[i][0]],'yaxis_title':renames[fig_cols[i][1]]}) for i in range(len(fig_cols))]
            for sn in metrics.SN.unique():
                df=metrics[metrics.SN==sn]
                for i,fig in enumerate(figs):
                    #cur_fig=px.scatter(df, x=renames[fig_cols[i][0]], y=renames[fig_cols[i][1]], text='SN')
                    marker={'size':12,'line':dict(width=2,color='DarkSlateGrey'),'symbol':'diamond'} if sn in highlights else {}
                    fig_data=go.Scatter(x=df[fig_cols[i][0]], y=df[fig_cols[i][1]], mode='markers+text',text=df['SN'],textposition="top center",marker=marker)
                    fig.add_trace(fig_data)

            return tuple(figs)

        figs=get_figs(highlights=['latest'])

        self.ptc_cpn_graph=dcc.Graph(figure=figs[0])
        self.ptc_sbr_cpn_sbr_graph=dcc.Graph(figure=figs[1])
        self.cpn_b_cpn_graph=dcc.Graph(figure=figs[2])
        self.ptc_b_ptc_graph=dcc.Graph(figure=figs[3])

        self.combiner=html.Div('Not ready',style={'display':'None'})

        self.layout=html.Div([
                            self.combiner,
                            dbc.Row([dbc.Col(self.display),dbc.Col(self.highlight)]),
                            dcc.Loading([
                                dbc.Row([dbc.Col(self.ptc_cpn_graph),dbc.Col(self.ptc_sbr_cpn_sbr_graph)]),
                                dbc.Row([dbc.Col(self.cpn_b_cpn_graph),dbc.Col(self.ptc_b_ptc_graph)])
                            ])
                        ])

        self.properties={'info':self.combiner.children}

        @self([self.ptc_cpn_graph.figure,self.ptc_sbr_cpn_sbr_graph.figure,self.cpn_b_cpn_graph.figure,self.ptc_b_ptc_graph.figure],[self.info,self.display.value,self.highlight.value])
        def update_plots(info,display,highlights):
            metrics=self.metrics.astype({'Date':'datetime64'})
            if 'all' not in display:
                metrics=metrics[metrics.SN.isin(display)]

            if not isinstance(highlights,list): highlights=[highlights]
            clean_highlights=[]
            for highlight in highlights:
                if highlight=='latest' and metrics.shape[0]>0:
                    clean_highlights.append(metrics.sort_values(by='Date')['SN'].to_list()[-1])
                elif highlight=='none':
                    pass
                elif highlight in metrics.SN.to_list():
                    clean_highlights.append(highlight)

            return get_figs(highlights=clean_highlights,metrics=metrics)



import dash
from .base import *
from .base_component import Component
from .uploader import Uploader
from .dropdown import Dropdown, add_to_dropdown
from ..utils import get_dynamic_values
import os

class File_Browser(Component):
    def __init__(self,label=None,multi=True,starting_folders=None, #list of paths or list of (label,path) [('Transfer','\\\\HOTH\\Transfer\\'),('Server','E:\\Shared Folders\\GTLDMS\\')]
            allowed_extensions=None,upload_to=None,keep_old_uploads=True):

        if label is None:
            label='Files' if multi else 'File'
        if starting_folders is None:
            if 'EXTERNAL_ROOT' in os.environ and os.path.isdir(os.environ['EXTERNAL_ROOT']):
                EXTERNAL_ROOT=os.environ['EXTERNAL_ROOT']
                starting_folders=[os.path.join(EXTERNAL_ROOT,x) for x in os.listdir(EXTERNAL_ROOT)]
            else:
                starting_folders=[]

        starting_folders_dict={}
        for fol in starting_folders:
            if isinstance(fol,str):
                starting_folders_dict[fol]=fol
            else:
                starting_folders_dict[fol[0]]=fol[1]

        self.path=dcc.Input(value='',className='w-100',disabled=True)
        self.folder_list=html.Div()
        self.folder_item=dbc.ListGroupItem

        children = [self.path,self.folder_list]

        self.dropdown=Dropdown(multi=multi,label=f'Selected {label}')

        if upload_to is not None:
            self.uploader=Uploader(type='area',multi=multi,label=label,allowed_extensions=allowed_extensions,upload_to=upload_to,dropdown=self.dropdown,keep_old_uploads=keep_old_uploads)
            children.append(self.uploader)

        children.append(self.dropdown)
        self.layout = html.Div(children)

        self.properties={'path':self.path.value,
                        'dir':self.folder_list.children,
                        'items':self.folder_item.children,
                        'value':self.dropdown.value,
                        'options':self.dropdown.options}

        #Base_callbacks
        #Show files in current folder
        get_list_item = lambda x,name=None: self.folder_item(x,index=x,style={'padding': '3px 10px','cursor': 'pointer'},className="list-group-item-action",n_clicks=0)
        @self(self.dir,self.path,self.dir)
        def show_folder_contents(path,folders):
            names=list(starting_folders_dict)
            if path is not None:
                if os.path.isfile(path):
                    return folders
                elif os.path.isdir(path):
                    names=['...']+[x for x in os.listdir(path) if allowed_extensions is None or any([x.lower().endswith(ext) for ext in allowed_extensions]) or os.path.isdir(os.path.join(path,x))]

            children=[get_list_item(x) for x in list(names)]
            return [dbc.ListGroup(children,style={"maxHeight": "300px", "overflow-y": "auto"})]

        #change dropdown
        @self([self.dropdown.options,self.dropdown.value],self.path,[self.dropdown.options,self.dropdown.value])
        def file_browser_add_to_dropdown(full_path,options,values):
            if os.path.isfile(full_path):
                new_values=[full_path]
                new_options = [{'label': os.path.split(full_path)[1], 'value': full_path}]
                return add_to_dropdown(new_options,new_values,options,values,multi=multi,keep_old_uploads=keep_old_uploads)
            return None

        #Change full-path text
        @self(self.path,self.folder_item.n_clicks,self.path)
        def change_full_path(inputs,full_path_root):
            if full_path_root is None: full_path_root=''
            if len(inputs)>0 and max(inputs)>0:
                index_vals=get_dynamic_values(self.folder_item.id[0]['type'])
                text=[index for index,val in index_vals.items() if val>0][0]
                full_path_root_isdir=os.path.isdir(full_path_root)
                if text in starting_folders_dict:
                    full_path_root = starting_folders_dict[text]
                elif full_path_root_isdir or os.path.isfile(full_path_root):
                    full_path=os.path.join(full_path_root,text)
                    if text=='...':
                        text=os.path.dirname(full_path_root if full_path_root_isdir else os.path.dirname(full_path_root))
                        folders=text.split(os.sep)
                        if text==full_path_root or not os.path.isdir(text) or (folders[0].lower().startswith('e:') and len(folders)<3):
                            text=''
                        full_path_root=text
                    elif os.path.isdir(full_path) or os.path.isfile(full_path):
                        full_path_root=full_path
                elif os.path.isdir(text):
                    full_path_root=text
            return full_path_root



import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_table
import pandas as pd
import numpy as np


#Main layout
layout = html.Div()



from .base_component import generate_all_callbacks, Component, Callback
from .alert import Alert
from .uploader import Uploader
from .dropdown import Dropdown,add_to_dropdown
from .file_browser import File_Browser
from .tabs import Tabs
from .download import Download
from .modal import Modal



import logging
import types
import os
import io

from .object_wrapper import put_object, get_object, list_objects, copy_object, delete_object, delete_objects, put_acl, get_acl, object_exists, open_object
from .bucket_wrapper import bucket_exists, get_buckets, get_bucket
from .utils import data_to_bytes

logger = logging.getLogger(__name__)

class S3Storage(object):
    def __init__(self,config,verify=True):
        self.region=config.AWS_DEFAULT_REGION
        self.bucket=self.get_bucket(config.BUCKET_NAME,verify=verify)
    def get_bucket(self,name,verify=True):
        if not verify or bucket_exists(name):
            return get_bucket(name,self.region)
        logger.warning("Verification of bucket '%s' failed",name)
        raise

    def get_buckets(self):
        return get_buckets()
    def exists(self,path):
        return object_exists(self.bucket,path)
    def open(self,path,mode='rb'):
        return open_object(self.bucket,path,mode=mode)
    def save(self,path,data):
        put_object(self.bucket,path,data)
    def delete(self,path):
        return delete_object(self.bucket,path)
    def list(self,prefix=None):
        return list_objects(self.bucket,prefix)

class LocalStorage(object):
    exists=lambda self,*args,**kwargs: os.path.exists(*args,**kwargs)
    open=lambda self,path,mode='r': io.open(path,mode)
    delete=lambda self,*args,**kwargs: os.remove(*args,**kwargs)
    list=lambda self,*args,**kwargs: os.listdir(*args,**kwargs)
    def save(self,path,data):
        data = data_to_bytes(data)
        with open(path,'wb') as f:
            f.write(data)



from pymongo.cursor import Cursor
from pymongo import MongoClient
import io
import pickle
import os

def get_client(url='l3-37:27017/',username=None,password=None,srv=False):
    prefix='mongodb'
    if srv: prefix+='+srv'
    if username and password:
        conn_str= f'{prefix}://{username}:{password}@{url}'
    else:
        conn_str=f'{prefix}://{url}'
    return MongoClient(conn_str)

def to_list(objs,field=None,query=None):
    if not isinstance(objs,Cursor):
        if query is None: query={}
        objs=objs.find(query)
    if field is None:
        return list(objs)
    return [getattr(x,field) if hasattr(x,field) else x[field] for x in objs]

def to_bytesio(obj):
    buffer=io.BytesIO()
    pos=buffer.tell()
    pickle.dump(obj,buffer,pickle.HIGHEST_PROTOCOL)
    buffer.seek(pos)
    return buffer

class BytesIOWrapper(io.BufferedReader):
    """Wrap a buffered bytes stream over TextIOBase string stream."""

    def __init__(self, text_io_buffer, encoding=None, errors=None, **kwargs):
        super(BytesIOWrapper, self).__init__(text_io_buffer, **kwargs)
        self.encoding = encoding or (text_io_buffer.encoding if hasattr(text_io_buffer,'encoding') else None) or 'utf-8'
        self.errors = errors or (text_io_buffer.errors if hasattr(text_io_buffer,'errors') else None) or 'strict'

    def _encoding_call(self, method_name, *args, **kwargs):
        raw_method = getattr(self.raw, method_name)
        val = raw_method(*args, **kwargs)
        return val.encode(self.encoding, errors=self.errors)

    def read(self, size=-1):
        return self._encoding_call('read', size)

    def read1(self, size=-1):
        return self._encoding_call('read1', size)

    def peek(self, size=-1):
        return self._encoding_call('peek', size)

class StringIOWrapper(BytesIOWrapper):
    """Wrap a buffered text stream over BytesIOBase stream."""

    def _encoding_call(self, method_name, *args, **kwargs):
        raw_method = getattr(self.raw, method_name)
        val = raw_method(*args, **kwargs)
        try:
            res=val.decode(self.encoding, errors=self.errors)
        except:
            res=val.decode('latin1', errors=self.errors)
        return res

def obj_to_buffer(obj):
    buffer=None
    info={}
    if isinstance(obj,str) and os.path.exists(obj):
        with open(obj,'rb') as f:
            buffer = io.BytesIO(f.read())
        filename=os.path.basename(obj)
        modified=os.path.getmtime(obj)
        info={'filename':filename,'modified':modified,'obj_type':'file'}
    elif isinstance(obj,io.BytesIO):
        obj.seek(0)
        buffer=obj
        info['obj_type']='bytesio'
    elif isinstance(obj,io.StringIO):
        obj.seek(0)
        buffer=BytesIOWrapper(obj)
        info['obj_type']='stringio'
    else:
        buffer=to_bytesio(obj)
        info['obj_type']='obj'
    return buffer,info

def buffer_to_obj(buffer,obj_type=None,**kwargs):
    obj_type=obj_type if obj_type is not None else (kwargs['obj_type'] if 'obj_type' in kwargs else None)
    if obj_type=='stringio':
        obj=StringIOWrapper(buffer)
    elif obj_type=='obj':
        obj=pickle.load(buffer)
    else:
        obj=buffer

    return obj



from .base import *
from .base_component import Component

class Modal(Component):
    def __init__(self,open_label='Open Modal',open_color='primary',header='Header',body='Modal Content',action_label='Ok',action_color='primary',show_cancel=True,action_fcn=None,small=False):

        self.open=html.A(open_label,href='#',className=f"ml-auto btn btn-{open_color}{' btn-sm' if small else ''}")
        self.modal_header=dbc.ModalHeader(header)
        self.modal_body=dbc.ModalBody(body)
        self.cancel=html.A('Cancel',href='#',className="ml-auto btn btn-secondary mr-3")
        self.action=html.A(action_label,href='#',className=f"ml-auto btn btn-{action_color}")
        self.layout = dbc.Modal([
                        self.modal_header,self.modal_body,
                        dbc.ModalFooter(dbc.Row(([self.cancel] if show_cancel else [])+[self.action],justify='end')),
                    ])

        self.properties={'is_open':self.layout.is_open,
                        'header':self.modal_header.children,
                        'body':self.modal_body.children}
        self.action_clicks=0
        @self(self.layout.is_open,[self.open.n_clicks,self.cancel.n_clicks,self.action.n_clicks],self.layout.is_open)
        def show_modal(open_click,cancel_click,action_click,is_open):
            if action_click!=self.action_clicks and action_fcn is not None:
                action_fcn()
                self.action_clicks=action_click
            if open_click or cancel_click or action_click:
                return not is_open
            return is_open



import dash
from flask import Flask
import dash_bootstrap_components as dbc
import dash_html_components as html
import os
from flask import request

from .components.base import *
from . import components
from .utils import open_browser
from .config import Config

def create_app(main_component,pathname_prefix='/',config=None):
    server = Flask(__name__)
    SECRET_KEY=config.SECRET_KEY if config is not None else 'adsfyiu3S3!jE%$axbjhwa195sxc@S'
    server.secret_key=SECRET_KEY

    app = dash.Dash(server=server,
        url_base_pathname=pathname_prefix,
        external_stylesheets=[dbc.themes.BOOTSTRAP],
        suppress_callback_exceptions=True)
    app.scripts.config.serve_locally=True
    app.css.config.serve_locally=True

    #Download URL
    components.Download.register(app,prefix=pathname_prefix)

    # Create Dash Layout
    main_app=main_component if isinstance(main_component,components.Component) else main_component(config)
    app.layout=main_app.generate_layout()
    components.generate_all_callbacks(app,verbose=0)
    return app

def start_app(main_component,pathname_prefix='/',debug=True, dev_tools_ui=True, dev_tools_props_check=True, host='0.0.0.0',port=5000,config=None):
    if config is None:
        config=Config()
    dashapp=create_app(main_component,pathname_prefix='/',config=config)
    app=dashapp.server
    dashapp.run_server(debug=debug, dev_tools_ui=dev_tools_ui, dev_tools_props_check=dev_tools_props_check, host=host,port=port)

def start_basic_server():
    #Start a basic dash server on port 5000
    class Main(components.Component):
        def layout(self):
            self.full_layout=html.Div('Test Dash server')
            return self.full_layout
    start_app(Main())



import tkinter as tk

class App(tk.Frame):
    def __init__(self, master=None):
        super().__init__(master)
        self.pack()

    def run(self):
        root = tk.Tk()
        app = Application(master=root)
        app.mainloop()

if __name__ == '__main__':
    # create the application
    myapp = App()

    #
    # here are method calls to the window manager class
    #
    myapp.master.title("My Do-Nothing Application")
    myapp.master.maxsize(1000, 400)
    myapp.mainloop()



from .base import *
from .base_component import Component
from ..utils import get_dynamic_values

class Alert(Component):
    def __init__(self,duration=5):
        self.num_alerts=0
        self.alert=dbc.Alert  #(children='',color='primary')
        self.remove_alert=html.A  #('X',className='border')
        #self.interval=dcc.Interval(interval=duration * 1000, n_intervals=0)
        #dbc.Collapse([self.alert,self.interval],is_open=False)

        self.layout=html.Div()

        self.properties={'alerts':self.layout.children}

        '''@self([self.layout.is_open,self.interval.disabled],self.msg)
        def show_hide_msg(msg):
            return (False,True) if msg is None or msg =='' else (True,False)'''

        @self(self.layout.children,self.remove_alert.n_clicks,self.layout.children)
        def reset_msg(n_intervals,alerts):
            index_vals=get_dynamic_values(self.remove_alert.id[0]['type'])
            index=[index for index,val in index_vals.items() if val>0][0]
            return ([x for x in alerts if x['props']['id']['index']!=index],)

    def create(self,msg,color='primary'):
        self.num_alerts+=1
        return self.alert(children=dbc.Row([msg,self.remove_alert('X',index=str(self.num_alerts),href='#',className=f"btn btn-sm btn-{color}")],justify="between"),index=str(self.num_alerts),color=color)



from IPython.display import display, HTML,Image,Javascript,FileLink
from ipywidgets import widgets, interact, interact_manual, interactive, AppLayout, Button, Layout, GridspecLayout
import os, threading, urllib, math
from queue import Queue

from ..notebook import get_notebook_outputs, get_notebook_inputs
from .widget_utils import *
from ..convert import docx_to_pdf, create_word_from_html
from ..utils import get_output_path, outputs_to_html, outputs_to_output, get_outputs_figure_widgets
from .record import get_previous_report, get_previous_inputs, record_report

#Variables
notebook_dir=os.path.join(os.getcwd(),'Types')
report_dir=os.path.join(os.getcwd(),'Reports')
template_dir=os.path.join(os.getcwd(),'Templates')
data_dir='/home/jovyan/external/Server/BTR/Uploaded'

#Helpers
debug_view = widgets.Output(layout={'border': '1px solid black'})
wrap=create_wrap(debug_view)

#Status valid
valid='<span style="color: green;">●</span> Ready'
busy='<span style="color: orange;">●</span> Busy'
invalid='<span style="color: red;">●</span> Required'
isvalid={True:valid,False:invalid,'busy':busy}

def get_outputs_threaded(path,inputs,progress,info):
    que = Queue()
    kwargs=dict(path=path,inputs=inputs,progress=progress,info=info)
    #thread = threading.Thread(target=run_notebook_as_code, kwargs=kwargs)
    thread = threading.Thread(target=lambda q, arg1: q.put(get_notebook_outputs(**kwargs)), args=(que, kwargs))
    thread.start()
    thread.join()
    return que.get()

def get_outputs(path,inputs,progress,info):
    return get_notebook_outputs(path,inputs=inputs,progress=progress,info=info)

#Callbacks
def get_browser_options(path='/home/jovyan',allow=None,start_path='/home/jovyan'):
    root,dirs,files=next(os.walk(path))
    res=[] if start_path==path else ['...']
    res.extend([x for x in dirs if not x.startswith('.')])
    if allow is not None and not isinstance(allow,list):
        allow=[allow]
    res.extend([x for x in files if allow is None or any([x.lower().endswith(i.lower()) for i in allow])])
    return res

ALLOW_BROWSER=True
def change_path(browser):
    def rtn(*args,**kwargs):
        global ALLOW_BROWSER
        if browser.value is None or not ALLOW_BROWSER:
            return
        ALLOW_BROWSER=False
        if os.path.isfile(browser.path):
            browser.path=os.path.dirname(browser.path)
        fullpath=os.path.join(browser.path,browser.value)
        if browser.value=='...':
            fullpath=os.path.dirname(browser.path)
            browser.options=get_browser_options(fullpath,allow=browser.allow,start_path=browser.start_path)
            browser.path=fullpath
            browser.rows=len(browser.options)
            browser.value=None
        elif os.path.isdir(fullpath):
            browser.options=get_browser_options(fullpath,allow=browser.allow,start_path=browser.start_path)
            browser.path=fullpath
            browser.rows=len(browser.options)
            browser.value=None
        elif os.path.isfile(fullpath):
            browser.path=fullpath
        ALLOW_BROWSER=True
    return rtn

def upload_db(uploader,upload_to=None):
    if upload_to is None:
        upload_to=data_dir
    def rtn(*args,**kwargs):
        for uploaded_file in uploader.value:
            uploader.path=os.join(upload_to,uploaded_file["name"])
            with open(uploader.path, "wb") as fp:
                fp.write(uploaded_file["content"])
    return rtn

INPUTS={}
REQUIRED_INPUTS=[]
def get_inputs():
    global INPUTS
    return {key:str(widget.value if not hasattr(widget,'path') else widget.path if isinstance(widget.allow,list) and any([widget.path.endswith(x) for x in widget.allow]) else None) for key,widget in INPUTS.items()}

def inputs_ready():
    global REQUIRED_INPUTS
    inputs=get_inputs()
    return all([val not in ['','None','none',None] for key,val in inputs.items() if key in REQUIRED_INPUTS])



def create_input_widgets(report,inputs,callback=None):
    def rtn(*args,**kwargs):
        global INPUTS
        global REQUIRED_INPUTS
        INPUTS={}
        path=os.path.join(notebook_dir,report.value)+'.ipynb'
        children=[]
        REQUIRED_INPUTS=[]

        CUSTOM_KEYS=['required','type','label','upload_to','description']

        if os.path.isfile(path):
            ins=get_notebook_inputs(path)
            #h_children=[]
            for key,val in ins.items():
                var_type=val['type']

                kwargs={'layout':Layout(width='auto')}
                kwargs.update({key:val for key,val in val.items() if key not in CUSTOM_KEYS})

                if 'required' in val and val['required']==True:
                    REQUIRED_INPUTS.append(key)

                if var_type=='autofill':
                    options=get_previous_inputs(notebook=path,inputs=[key])
                    kwargs.update(dict(options=options,ensure_option=False,disabled=False))
                    widget=widgets.Combobox(**kwargs)
                elif var_type=='upload':
                    widget=widgets.FileUpload(**kwargs)
                    widget.path=None
                    upload_to=val['upload_to'] if 'upload_to' in val else None
                    wrap(widget)(upload_db(widget,upload_to))
                elif var_type=='path':
                    widget=widgets.Select(**kwargs)
                    widget.path=val['path'] if 'path' in val else '/home/jovyan'
                    widget.start_path=val['start_path'] if 'start_path' in val else '/home/jovyan'
                    widget.allow=val['allow'] if 'allow' in val else None
                    widget.options=get_browser_options(path=widget.path,allow=widget.allow,start_path=widget.start_path)
                    widget.value=None
                    wrap(widget)(change_path(widget))
                elif hasattr(widgets,var_type):
                    widget=getattr(widgets,var_type)(**kwargs)
                else:
                    continue

                #new_children=[]
                label=val['label'] if 'label' in val else key
                children.append(widgets.HTML(label+':'))
                if 'description' in val:
                    children.append(widgets.HTML(f'<p style="color:#545b62;">{val["description"]}</p>'))
                children.append(widget)
                #children.append(widgets.VBox(children=new_children,layout=Layout(width='100%')))

                #if len(h_children)==2:
                #    children.append(widgets.HBox(h_children,layout=Layout(width='auto')))
                #    h_children=[]
                if callback is not None:
                    wrap(widget)(callback(widget))

                INPUTS[key]=widget

        #children.extend(h_children)
        inputs.children=children
    return rtn

def show_results(paths,info,download):
    info.value='Complete! You can download your report here:'
    res=[f'<p><a href="/user-redirect/files/{urllib.parse.quote(path.replace("/home/jovyan",""))}" target="_blank">{os.path.split(path)[-1]}</a></p>' for path in paths]
    download.value=(''.join(res))

OUTPUTS=[]

def run_code_fcn(report,exec_code_info,progress,info,notebook_output,edit_notebook,edit_figs,generate_report):
    @debug_view.capture(clear_output=True)
    def rtn(*_,notebook_path=None,inputs=None):
        global OUTPUTS
        if notebook_path is None:
            notebook_path=os.path.join(notebook_dir,f'{report.value}.ipynb')
        if inputs is None:
            inputs=get_inputs()
        #Execute code
        exec_code_info.children=[widgets.HTML(isvalid['busy']),progress]
        progress.value=.0
        progress.layout.display = "flex"
        print(notebook_path, inputs)
        OUTPUTS=get_outputs(path=notebook_path,inputs=inputs,progress=progress,info=info)
        exec_code_info.children=[widgets.HTML(isvalid[True])]
        edit_notebook.disabled = False
        edit_figs.disabled = False
        generate_report.disabled = False
        notebook_output.clear_output()
        progress.layout.display = "none"
    return rtn

def generate_report_fcn(report,docx_gen_info,pdf_gen_info,template,generate_pdf,info,download):
    def rtn(*_,notebook_path=None,inputs=None,display_results=True):
        global OUTPUTS
        if notebook_path is None:
            notebook_path=os.path.join(notebook_dir,f'{report.value}.ipynb')
        if inputs is None:
            inputs=get_inputs()
        #Generate report
        docx_gen_info.children=[widgets.HTML(isvalid['busy'])]
        info.value='Generating report...'
        download.value=''
        output_path=get_output_path(notebook_path,inputs=inputs,folder=report_dir,extension='docx')
        template_path=os.path.join(template_dir,f'{template.value}.docx')
        html=outputs_to_html(OUTPUTS,types=['text/html','application/vnd.jupyter.widget-view+json'])
        paths=create_word_from_html(html,template_path,output_path=output_path)
        docx_gen_info.children=[widgets.HTML(isvalid[True])]

        #Generate pdf
        if generate_pdf.value:
            info.value='Generating PDF...'
            pdf_gen_info.children=[widgets.HTML(isvalid['busy'])]
            paths.extend(docx_to_pdf(paths))
            pdf_gen_info.children=[widgets.HTML(isvalid[True])]

        record_report(notebook_path=notebook_path,inputs=inputs,paths=paths)
        if display_results:
            show_results(paths,info,download)
        return paths
    return rtn

def reset_exec_report_status(exec_code_info,docx_gen_info,pdf_gen_info,edit_notebook,edit_figs,generate_report,notebook_output):
    def rtn(*_):
        exec_code_info.children=[widgets.HTML(isvalid[False])]
        docx_gen_info.children=[widgets.HTML(isvalid[False])]
        pdf_gen_info.children=[widgets.HTML(isvalid[False])]
        edit_notebook.disabled = True
        edit_figs.disabled = True
        generate_report.disabled = True
        notebook_output.clear_output()
    return rtn

def reset_and_run_code(report,progress,info,exec_code_info,docx_gen_info,pdf_gen_info,edit_notebook,edit_figs,generate_report,notebook_output):
    def rtn(*_):
        reset_exec_report_status(exec_code_info,docx_gen_info,pdf_gen_info,edit_notebook,edit_figs,generate_report,notebook_output)()
        run_code_fcn(report,exec_code_info,progress,info,notebook_output,edit_notebook,edit_figs,generate_report)()
    return rtn

def run_all(report,template,progress,info,download,force_execution,generate_pdf,exec_code_info,docx_gen_info,pdf_gen_info,
            edit_notebook,edit_figs,generate_report,notebook_output):
    def rtn(*_):
        #reset status
        reset_exec_report_status(exec_code_info,docx_gen_info,pdf_gen_info,edit_notebook,edit_figs,generate_report,notebook_output)()
        notebook_path=os.path.join(notebook_dir,f'{report.value}.ipynb')
        inputs=get_inputs()

        #Check for previous Reports
        paths=get_previous_report(notebook_path,inputs)
        if force_execution.value or paths is None:
            #Execute code
            run_code_fcn(report,exec_code_info,progress,info,notebook_output,edit_notebook,edit_figs,generate_report)(notebook_path=notebook_path,inputs=inputs)

            #Generate report
            paths=generate_report_fcn(report,docx_gen_info,pdf_gen_info,template,generate_pdf,info,download)(notebook_path=notebook_path,inputs=inputs,display_results=False)

        show_results(paths,info,download)
    return rtn

#Status checks
def set_generate_button_disabled(generate,options_info,inputs_info,run_code):
    def rtn():
        generate.disabled = run_code.disabled = not ((options_info.valid if hasattr(options_info,'valid') else False) and (inputs_info.valid if hasattr(inputs_info,'valid') else False))
    return rtn

def check_options(report,options_info,complete_fcn=None):
    def rtn(*_):
        valid=report.value !='' and report.value is not None
        options_info.children=[widgets.HTML(isvalid[valid])]
        options_info.valid=valid
        if complete_fcn is not None:
            complete_fcn()
    return rtn

def check_inputs(inputs_info,complete_fcn=None):
    def rtn(widget):
        def sub_rtn(*_):
            valid=inputs_ready()
            inputs_info.children=[widgets.HTML(isvalid[valid])]
            inputs_info.valid=valid
            if complete_fcn is not None:
                complete_fcn()
        return sub_rtn
    return rtn

def show_pdf_info(generate_pdf,pdf_info_box):
    def rtn(*_):
        pdf_info_box.layout.display = 'flex' if generate_pdf.value else 'none'
    return rtn

def add_report_to_output(output):
    def rtn(*_):
        global OUTPUTS
        output.clear_output()
        output.append_display_data(HTML('Collecting report outputs...'))
        output.clear_output()
        for out in OUTPUTS:
            if 'text/html' in out.data:
                out.data['text/html']='<div contenteditable="false">'+out.data['text/html']+'</div>'
        outputs_to_output(OUTPUTS,output)
    return rtn

def create_limit_widget(ax,axis_type):
    lim=ax.get_xlim() if axis_type=='x' else ax.get_ylim()
    diff=abs(lim[1]-lim[0])
    readout_format=f'.{int(max([0,math.log10(10/diff)]))}f'
    desc='X-axis:' if axis_type=='x' else 'Y-axis:'
    widget=widgets.FloatRangeSlider(value=lim,min=lim[0]-(diff*0.1),max=lim[1]+(diff*0.1),step=diff/100,description=desc,disabled=False,continuous_update=False,orientation='horizontal',readout=True,readout_format=readout_format)

    def rtn(*_):
        if axis_type=='y':
            ax.set_ylim(widget.value)
        elif axis_type=='x':
            ax.set_xlim(widget.value)

    wrap(widget)(rtn)
    return widget

def create_legend_widget(ax):
    if any([x.get_label() not in [None, '', ' '] for x in ax.lines]) and len(ax.lines)>1:
        options=['best','upper right','upper left','lower left','lower right','right','center left','center right','lower center','upper center','center']
        widget=widgets.Dropdown(value='best',options=options,rows=5,layout=Layout(width='120px'))

        def rtn(*_):
            ax.legend(loc=widget.value)

        wrap(widget)(rtn)
        return [widgets.HTML('Legend:'),widget]
    return []

def add_figs_to_output(output):
    def rtn(*_):
        global OUTPUTS
        output.clear_output()
        output.append_display_data(HTML('Collecting figures...'))
        canvases=get_outputs_figure_widgets(OUTPUTS)

        res=[]
        complete_ax=[]
        def get_title(ax):
            title=ax.get_title()
            return f'{title+" - " if title not in [None,""," "] else ""}{ax.get_ylabel()}'
        for canvas in canvases:
            children=[]
            len_axes=len(canvas.figure.axes)
            for i,ax in enumerate(canvas.figure.axes):
                if ax in complete_ax:
                    continue
                shared=list(ax.get_shared_x_axes())
                shared_x_axes=[shared_ax for x in shared if ax in x for shared_ax in x]
                shared=list(ax.get_shared_y_axes())
                shared_y_axes=[shared_ax for x in shared if ax in x for shared_ax in x]

                if len(shared_x_axes)>0:
                    res.append(widgets.HBox([widgets.HTML(f' ',layout=Layout(width='200px')),create_limit_widget(ax,axis_type='x')],layout=Layout(width='auto')))
                    for shared_ax in shared_x_axes:
                        res.append(widgets.HBox([widgets.HTML(f'<strong>{get_title(shared_ax)}:</strong>',layout=Layout(width='200px')),create_limit_widget(shared_ax,axis_type='y'),*create_legend_widget(shared_ax)],layout=Layout(width='auto')))
                        complete_ax.append(shared_ax)
                elif len(shared_y_axes)>0:
                    res.append(widgets.HBox([widgets.HTML(f' ',layout=Layout(width='200px')),create_limit_widget(ax,axis_type='y')],layout=Layout(width='auto')))
                    for shared_ax in shared_y_axes:
                        res.append(widgets.HBox([widgets.HTML(f'<strong>{get_title(shared_ax)}:</strong>',layout=Layout(width='200px')),create_limit_widget(shared_ax,axis_type='x'),*create_legend_widget(shared_ax)],layout=Layout(width='auto')))
                        complete_ax.append(shared_ax)
                else:
                    res.append(widgets.HBox([widgets.HTML(f'<strong>{get_title(ax)}:</strong>',layout=Layout(width='200px')),create_limit_widget(ax,axis_type='x'),create_limit_widget(ax,axis_type='y'),*create_legend_widget(ax)],layout=Layout(width='auto')))
                    complete_ax.append(ax)

            res.append(widgets.Box([canvas],layout=Layout(width='100%',border="1px solid grey")))
            #canvas.manager.toolbar._Button("PREVIOUS", "back_large", self.prev)
        output.clear_output()
        outputs_to_output(res,output)
    return rtn


# CREATE APP
def create_app():
    inputs=widgets.VBox(layout=Layout(width='100%'))
    notebook_options=[x[:-6] for x in os.listdir(notebook_dir) if x.endswith('.ipynb')]
    val='Snapshot' if 'Snapshot' in notebook_options else notebook_options[0]
    report=widgets.Dropdown(options=notebook_options,rows=5,value=val,layout=Layout(width='auto'))
    template_options=['None']+[x[:-5] for x in os.listdir(template_dir) if x.endswith('.docx')]
    template_val='template' if 'template' in template_options else template_options[0]
    template=widgets.Dropdown(options=template_options,rows=5,value=template_val,layout=Layout(margin='0px 30px 0px 0px'))
    generate_pdf=widgets.Checkbox(value=False,description='Create PDF',disabled=False,indent=False)
    force_execution=widgets.Checkbox(value=True,description='Force Execution',disabled=False,indent=False)

    options=widgets.HBox([widgets.HTML('Template:'),template,generate_pdf,force_execution],layout=Layout(width='100%',padding='10px 0px 0px 0px'))

    #Status
    progress = widgets.FloatProgress(value=0.0, min=0.0, max=1.0)
    progress.layout.display = "none"

    options_info=widgets.HBox([widgets.HTML(invalid)],layout=Layout(width='auto'))
    inputs_info=widgets.HBox([widgets.HTML(invalid)],layout=Layout(width='auto'))
    exec_code_info=widgets.HBox([widgets.HTML(invalid)],layout=Layout(width='auto'))
    docx_gen_info=widgets.HBox([widgets.HTML(invalid)],layout=Layout(width='auto'))
    pdf_gen_info=widgets.HBox([widgets.HTML(invalid)],layout=Layout(width='auto'))

    pdf_info_box=widgets.HBox([widgets.HTML('<strong>PDF Generation:</strong>'),pdf_gen_info],layout=Layout(width='auto'))
    pdf_info_box.layout.display = "none"
    status=widgets.VBox([
        widgets.HBox([widgets.HTML('<strong>Report Options:</strong>'),options_info],layout=Layout(width='auto')),
        widgets.HBox([widgets.HTML('<strong>Report Inputs:</strong>'),inputs_info],layout=Layout(width='auto')),
        widgets.HBox([widgets.HTML('<strong>Code Execution:</strong>'),exec_code_info],layout=Layout(width='auto')),
        widgets.HBox([widgets.HTML('<strong>Document Generation:</strong>'),docx_gen_info],layout=Layout(width='auto')),
        pdf_info_box
    ],layout=Layout(width='500px'))


    info=widgets.HTML()
    download=widgets.HTML()
    generate=widgets.Button(description='Run Code and Generate Report',disabled=True,button_style='primary', icon='cogs',layout=Layout(width='auto'))
    results=widgets.VBox([generate,info,download],layout=Layout(width='auto'))

    top_area=widgets.HBox([status,results],layout=Layout(width='auto'))

    #Edit
    run_code=widgets.Button(description='Run Code',disabled=True,button_style='primary', icon='play')
    generate_report=widgets.Button(description='Generate Report',disabled=True,button_style='primary', icon='cogs')
    edit_notebook=widgets.Button(description='Edit Report',disabled=True,button_style='primary', icon='file-text-o')
    edit_figs=widgets.Button(description='Edit Figures',disabled=True,button_style='primary', icon='line-chart')
    notebook_output=widgets.Output(layout={'border': '1px solid black'})
    notebook_output.append_display_data(HTML('Please run code first'))

    #Callbacks
    #Change inputs
    generate_disabled=set_generate_button_disabled(generate,options_info,inputs_info,run_code)
    callback=check_inputs(inputs_info,complete_fcn=generate_disabled)
    fcn=create_input_widgets(report,inputs,callback=callback)
    wrap(report)(fcn)
    fcn()


    fcn=check_options(report,options_info,complete_fcn=generate_disabled)
    wrap(report)(fcn)
    fcn()

    fcn=show_pdf_info(generate_pdf,pdf_info_box)
    wrap(generate_pdf)(fcn)
    fcn()

    run_code.on_click(reset_and_run_code(report,progress,info,exec_code_info,docx_gen_info,pdf_gen_info,edit_notebook,edit_figs,generate_report,notebook_output))
    edit_notebook.on_click(add_report_to_output(notebook_output))
    edit_figs.on_click(add_figs_to_output(notebook_output))
    generate_report.on_click(generate_report_fcn(report,docx_gen_info,pdf_gen_info,template,generate_pdf,info,download))

    generate.on_click(run_all(report,template,progress,info,download,force_execution,generate_pdf,exec_code_info,docx_gen_info,pdf_gen_info,
                edit_notebook,edit_figs,generate_report,notebook_output))

    next_button=widgets.Button(description='Next',disabled=False,button_style='primary', icon='arrow-right')
    content = widgets.Tab(children=[
                widgets.VBox([report,options,next_button],layout=Layout(width='auto')),
                widgets.VBox([inputs,next_button],layout=Layout(width='auto')),
                widgets.VBox([widgets.HBox([run_code,edit_notebook,edit_figs,generate_report],layout=Layout(width='auto')),notebook_output],layout=Layout(width='auto'))
            ], layout=Layout(width='auto',border='1px solid grey'))

    titles={0:'Options',1:'Inputs',2:'Edit'}
    for i,val in titles.items():
        content.set_title(i, val)
    content.selected_index=0

    def go_next(*_):
        content.selected_index+=1
    next_button.on_click(go_next)

    app=widgets.VBox([widgets.HTML('<h1>Generate Report</h1>'),top_area,content],
            layout=Layout(width='100%'))
    return app



from .db import *
from .classes import *

get_oracle_conn = lambda user,password,host,port,server: f'oracle+cx_oracle://{user}:{password}@{host}:{port}/{server}'
oracle_connections={'repower':dict(user='tracker_repower',password='tracker_repower',host='leia',port='1522',server='ora112'),
                    'tracker':dict(user='tracker_web',password='tracker_web',host='leia',port='1522',server='ora112'),
                    'oilat':dict(user='tracker_read',password='readonly',host='dooku',port='1521',server='nlsvc112')}
oracle_schemas={'oilat':'tracker_pwc',
                'tracker':'tracker_web',
                'repower':'tracker_repower',
                }

#Build-types-makes
def get_build_type_filters(tables):
    renames={'short_desc':'type','nomenclature':'make','description':'description','assembly_name':'model'}
    cols= [
        tables.slot.columns.slot_id,
        tables.slot.columns.nomenclature,
        tables.cdtbl_slot_class.columns.short_desc,
        tables.assembly.columns.assembly_name,
        tables.assembly.columns.description
    ]
    filters=[
        tables.slot.columns.assembly_id==tables.assembly.columns.assembly_id,
        tables.cdtbl_slot_class.columns.slot_class_cd==tables.slot.columns.slot_class_cd
    ]
    return cols,filters,renames

#organization functions
def organization_props(self):
    return {'name':self.organization_name,
        'type':self.get_org_type_cd[0].short_desc,
        'location_name':self.location_name,
        'street_address':self.street_address,
        'city':self.city,
        'state_province':self.state_province,
        'country':self.country,
        'postal_code':self.postal_code}

def organization_items(self,tables):
    return self.search(tables.inv_item).filter(
        tables.inv_item.columns.org_id==self.org_id,
        tables.inv_item.columns.sn==tables.inv_item.columns.h_sn,
        tables.inv_item.columns.nh_sn==None
    )

def org_parents(self,tables):
    return self.search(tables.organization).filter(
        tables.organization.columns.nh_org_id==self.org_id
    )

def organization_site_items(self,tables):
    return self.sess.query(
        tables.organization.organization_name,
        tables.inv_item
    ).filter(
        tables.inv_item.columns.org_id==tables.organization.org_id,
        tables.organization.columns.nh_org_id==self.org_id
    )

#inv_item functions
def inv_item_props(self):
    part_x_slot=self.get_slot_id[0]
    slot=part_x_slot.get_slot_id[0] # inv_item->part_x_slot->slot_id
    part=part_x_slot.get_part_id[0] # inv_item->part_x_slot->part_id
    return {'name':self.user_serial_no,
        'build':slot.build,
        'manufacturer':part.mfr_name,
        'manufacturer_part_no':part.mfr_part_no,
        'description':part.nomenclature}

def inv_item_data(self,tables):
    sc_sample=orm.aliased(tables.samp_collect)
    sc_test=orm.aliased(tables.samp_collect)
    return self.sess.query(
            tables.inv_item.columns.user_serial_no,
            sc_sample.columns.sample_cd,
            sc_sample.columns.sample_date,
            tables.pdm_samp_type.columns.samp_name,
            tables.data_collect_result.columns.data_value,
            tables.cdtbl_analy_method.columns.short_desc,
            tables.cond_ind.columns.ind_name,
            tables.cond_ind.columns.ind_desc,
            tables.cond_ind.columns.ind_symbol
        ).filter(
            tables.inv_item.columns.sn==self.sn,
            sc_sample.columns.delete_flg == 'N',
            sc_sample.columns.sn==self.sn,
            sc_sample.columns.samp_ident_no==sc_sample.columns.h_samp_ident_no,
            sc_test.columns.h_samp_ident_no==sc_sample.columns.samp_ident_no,
            sc_test.columns.samp_ident_no != sc_sample.columns.samp_ident_no,
            sc_test.columns.samp_type_no == tables.pdm_samp_type.columns.samp_type_no,
            sc_test.columns.samp_ident_no == tables.data_collect_result.columns.samp_ident_no,
            tables.data_collect_result.columns.ind_no == tables.cond_ind.columns.ind_no,
            tables.cdtbl_analy_method.columns.analy_method_cd==tables.data_collect_result.columns.analy_method_cd
        ).order_by(sc_sample.columns.sample_date.asc())

def inv_item_limits(self,tables):
    return self.sess.query(
            tables.slot.columns.slot_id,
            tables.slot.columns.nomenclature,
            tables.limit_def.columns.lower_limit,
            tables.cond_ind_status_def.columns.status_name,
            tables.cond_ind.columns.ind_name,
            tables.cond_ind.columns.ind_symbol,
            tables.cond_ind.columns.ind_desc
        ).filter(
            tables.inv_item.columns.sn==self.sn,
            tables.slot.columns.slot_id==tables.inv_item.columns.slot_id,
            tables.limit_def.columns.object_id==tables.slot.columns.slot_id,
            tables.cond_ind.columns.ind_no==tables.pdm_samp_cond_ind.columns.ind_no,
            tables.pdm_samp_cond_ind.columns.ind_no==tables.cond_ind_status_def.columns.ind_no,
            tables.cond_ind_status_def.columns.status_no==tables.limit_def.columns.status_no
        )

def inv_item_components(self,tables):
    return self.search(
        tables.inv_item
    ).filter(
        tables.inv_item.columns.sn==tables.inv_item.columns.h_sn,
        tables.inv_item.columns.nh_sn==self.sn
    )

def inv_item_other_items(self,tables):
    return self.search(
        tables.inv_item
    ).filter(
        tables.inv_item.columns.h_sn==self.sn,
        tables.inv_item.columns.nh_sn==self.sn
    )

def inv_item_sensors(self,tables):
    return self.search(
        tables.inv_item
    ).filter(
        tables.inv_item.columns.h_sn==self.sn,
        tables.inv_item.columns.nh_sn==self.sn,
        tables.slot.columns.slot_class_cd=='LN',
        tables.inv_item.columns.slot_id==tables.part_x_slot.columns.slot_id,
        tables.part_x_slot.columns.slot_id==tables.slot.columns.slot_id
    ).distinct()

#slot
def slot_build_type(self,tables):
    cols,filters,renames=get_build_type_filters(tables)
    filters.append(tables.slot.columns.slot_id==self.slot_id)
    builds=self.sess.query(*cols).filter(*filters).all()
    res=[{rename:getattr(build,key) for key,rename in renames.items()} for build in builds]
    return res[0] if len(res)==1 else res

def get_table_attribs(tables):
    return {'organization':{
            'props':organization_props,
            'sites':lambda self: self.search(tables.organization).filter(tables.organization.columns.nh_org_id==self.org_id).distinct(),
            'parents':lambda self: org_parents(self,tables=tables),
            'items':lambda self: organization_items(self,tables=tables),
            'site_items':lambda self: organization_site_items(self,tables=tables)
        },
        'inv_item':{
            'organization':lambda self: self.get_org_id[0],
            'props': inv_item_props,
            'data':lambda self: query_to_df(inv_item_data(self,tables=tables)),
            'limits': lambda self: inv_item_limits(self,tables=tables),
            'components':lambda self: inv_item_components(self,tables=tables),
            'other':lambda self: inv_item_other_items(self,tables=tables),
            'sensors':lambda self: inv_item_sensors(self,tables=tables)
        },
        'slot':{
            'build': lambda self: slot_build_type(self,tables=tables),
        }
    }

class Analyzer(DB):
    def __init__(self,db_params=None,schema=None):
        if isinstance(db_params,str):
            assert db_params in oracle_connections, f"{db_params} is not available in standard definitions, please provide the user, password, host, port and server instead."
            if schema is None and db_params in oracle_schemas:
                schema=oracle_schemas[db_params]
            db_params=oracle_connections[db_params]
        assert isinstance(db_params,dict), f"A dictionary of the databe connection is required. Instead recieved db as {type(db_params)}"
        conn_params=['user', 'password', 'host', 'port', 'server']
        missing_params=[x for x in conn_params if x not in db_params]
        assert len(missing_params)==0, f"Missing connection parameters: {', '.join(missing_params)}"
        self.db_params={key:val for key,val in db_params.items() if key in conn_params}
        conn_str=get_oracle_conn(**self.db_params)
        if schema is None:
            schema=self.db_params['user']
        super().__init__(conn_str,schema=schema,
                        table_names=['cond_ind','slot','part','assembly','data_collect_result','inv_item','organization','samp_collect',
                                    'pdm_samp_type','contacts','cond_ind_status_def','limit_def','cdtbl_org_type','cdtbl_unit_of_meas',
                                    'cdtbl_meas_type','pdm_samp_cond_ind','cdtbl_analy_method','cdtbl_present_format','part_x_slot','cdtbl_object_type'
        ])
    def __str__(self):
        connected="connected" if hasattr(self,'sess') else "not connected"
        return f'{self.db_params["server"]} on {self.db_params["host"]} with schema {self.schema} ({connected})'
    def connect(self,*args,**kwargs):
        self=super().connect(*args,**kwargs)

        extra_attribs=get_table_attribs(self.tables)
        for table_name,attribs in extra_attribs.items():
            self.tables.__dict__[table_name].add_result_attribs(attribs)
        return self
    @property
    def props(self):
        if hasattr(self,'sess'):
            return {
                'orgs':self.tables.organization.objects.count(),
                'items':self.tables.inv_item.objects.count()
            }
        else:
            return 'Not connected'
    @property
    def fields(self):
        return self.sess.query(
                self.tables.pdm_samp_type.columns.samp_name,
                self.tables.cond_ind.columns.ind_name,
                self.tables.cond_ind.columns.ind_desc,
                self.tables.cond_ind.columns.ind_symbol
            ).filter(
                self.tables.samp_collect.columns.samp_type_no == self.tables.pdm_samp_type.columns.samp_type_no,
                self.tables.samp_collect.columns.samp_ident_no == self.tables.data_collect_result.columns.samp_ident_no,
                self.tables.data_collect_result.columns.ind_no == self.tables.cond_ind.columns.ind_no
            ).distinct()
    @property
    def fields_df(self):
        return query_to_df(self.fields)
    def all_limits(self):
        tables=self.tables
        cols,filters,renames=get_build_type_filters(tables)
        cols=[
            tables.limit_def.columns.lower_limit,
            tables.cdtbl_object_type.columns.object_type_name,
            tables.limit_def.columns.is_single_value_flg,
            tables.cond_ind_status_def.columns.status_name,
            tables.cond_ind.columns.ind_name,
            tables.cond_ind.columns.ind_symbol,
            tables.cond_ind.columns.ind_desc
        ]+cols
        filters=[
            tables.slot.columns.slot_id==tables.limit_def.columns.object_id,
            tables.cdtbl_object_type.columns.object_type==tables.limit_def.columns.object_type,
            tables.cond_ind.columns.ind_no==tables.pdm_samp_cond_ind.columns.ind_no,
            tables.pdm_samp_cond_ind.columns.ind_no==tables.cond_ind_status_def.columns.ind_no,
            tables.cond_ind_status_def.columns.status_no==tables.limit_def.columns.status_no
        ]+filters
        return self.sess.query(*cols).filter(*filters).distinct()
    @property
    def limits(self):
        _,_,renames=get_build_type_filters(self.tables)
        df=query_to_df(self.all_limits())
        df=df.rename(columns=renames)
        build_cols=list(renames.values())
        df=df.set_index(build_cols).reorder_levels(build_cols).sort_index()
        return df



import re, os
import numpy as np
import pandas as pd
from itertools import groupby
from io import BytesIO, StringIO

def get_buffer_metas(buffers,split_buffer_meta=False,*args,**kwargs):
    if type(buffers) not in [list,tuple] or (len(buffers)==2 and isinstance(buffers[1],dict)):
        buffers=[buffers]
    res=[]
    kwargs.update(dict(pair for d in args if isinstance(d,dict) for pair in d.items()))
    for buffer in buffers:
        cur_meta=kwargs.copy()
        if type(buffer) in [list,tuple]:
            if len(buffer)>1:
                cur_meta.update(dict(pair for d in buffer[1:] for pair in d.items()))
            buffer=buffer[0]
        if hasattr(buffer,'_file'):
            cur_meta.update(buffer._file)
        res.append([buffer,cur_meta])
    if split_buffer_meta:
        return list(zip(*res))
    else:
        return res

def dfs_to_buffers(dfs):
    df_metas=get_buffer_metas(dfs)
    res=[]
    for df,meta in df_metas:
        #sort dfs
        if 'Time' in df.columns:
            df=df.sort_values(by=['Time'])
        elif 'RecordID' in df.columns:
            df=df.sort_values(by=['RecordID'])

        buffer = BytesIO()
        buffer.close=lambda: None
        #data=bytes(df.to_csv(index=False).encode("utf-8"))
        #buffer.write(data)
        data=df.to_pickle(buffer)

        buffer.seek(0)
        res.append([buffer,meta])
    return res

def df_to_buffer(df):
    return dfs_to_buffers(df)[0]

def buffers_to_dfs(buffers,single_df=False,**groupmeta):
    res=[]
    for buffer,meta in get_buffer_metas(buffers,**groupmeta):
        #file=StringIO()
        buffer.seek(0)
        #file.write(buffer.read().decode())
        #file.seek(0)
        #df=pd.read_csv(file)
        df=pd.read_pickle(buffer)
        res.append([df,meta])

    if single_df:
        dfs=[]
        for df,meta in res:
            dfs.append(df.assign(**{key:(', '.join([str(x) for x in val]) if isinstance(val,list) else val) for key,val in meta.items()}))
        res=pd.concat(dfs)
    return res

def buffer_to_df(buffer,**groupmeta):
    return buffers_to_dfs(buffer,single_df=True,**groupmeta)



import pandas as pd
import numpy as np
from urllib.parse import quote as urlquote

from .. import components
from ..components.base import *
from ..config import *
from ..utils import spectra_and_baseline

class Amplitudes(components.Component):
    def __init__(self,main_app):
        self.properties={}
        self.main_app=main_app
    def layout(self):
        self.alert=components.Alert()
        self.location=dcc.Location(id='url',refresh=False)

        self.selected_spectra=components.File_Browser(label='Spectra',multi=True,
                allowed_extensions=SPECTRA_EXTENSIONS,upload_to=SPECTRA_PATH,keep_old_uploads=True)
        self.background=dcc.Checklist(options=[{'label': 'Remove background', 'value': 'rm'}],value=[],className="p-2")

        # Configuration and regions
        default_config,default_value,config_options,configs=self.load_configs()
        self.cur_regions=pd.DataFrame({},columns=CONFIG_COLUMNS) if default_value is None else pd.read_csv(default_value)

        self.config_dropdown=components.Dropdown(value=default_value,options=config_options,multi=False,label=None)

        def check_config_uploads(file_paths):
            msg='';color='success'
            res={}
            for filename,path in file_paths.items():
                df=pd.read_csv(path)
                if set(df.columns)!=set(CONFIG_COLUMNS):
                    msg=f'{filename} did not have the correct columns, which should be: {", ".join(CONFIG_COLUMNS)}'
                    color='danger'
                else:
                    res[filename]=path
            return msg,color,res

        self.config_upload=components.Uploader(type='button',multi=False,label='Config',style=None,allowed_extensions=['.csv'],upload_to=CONFIG_PATH,dropdown=self.config_dropdown,keep_old_uploads=True,condition_check=check_config_uploads,alert=self.alert)
        self.config_upload.layout.obj.style={'display':'inline-block'}

        self.download_config = components.Download(filename=default_config,data=self.cur_regions).className('ml-2 mr-2')
        self.config_save_label=html.Label('Save configuration as:')
        self.config_save_input=dcc.Input(value=default_config,className='w-100')
        self.config_save_modal=components.Modal(open_label='Save As',header=f'Save {default_config}',body=html.Div([self.config_save_label,self.config_save_input]),action_label='Save',show_cancel=True)

        self.regions=dash_table.DataTable(columns=[{"name": i, "id": i} for i in self.cur_regions.columns],data=self.cur_regions.to_dict('records'),editable=True,
                                row_selectable='multi',selected_rows=list(range(self.cur_regions.shape[0])),css=[{'selector': '.row', 'rule': 'margin: 0'}])
        config_tools=html.Div([self.config_save_modal,
                            dbc.Row([dbc.Col(self.config_dropdown,width=4),dbc.Col([self.config_save_modal.open,self.download_config,self.config_upload])],
                            className='pb-2',justify="start")])

        #Results
        self.graph=dcc.Graph(figure=go.Figure(layout={'height':700,'xaxis_title':'Wavelength (µm)','yaxis_title':'Intensity'}))
        self.results_table_download=components.Download(filename='Results.csv',small=True)
        self.results_table=dash_table.DataTable(editable=False,css=[{'selector': '.row', 'rule': 'margin: 0'}])
        self.result_card=components.Tabs({'Spectra Chart':dcc.Loading(self.graph),
                                        'Table Results':dcc.Loading([
                                            dbc.Row(self.results_table_download,justify="start",className='pb-2'),
                                            dbc.Row(self.results_table)
                                        ])})

        self.combiner=html.Div('Not ready',style={'display':'None'})
        self.full_layout=html.Div([self.combiner,
                self.alert,
                self.location,
                dbc.Row([
                    dbc.Col(html.Div([self.selected_spectra,self.background]),width=4),
                    dbc.Col(html.Div([config_tools,html.Div(self.regions)]))
                ]),
                dbc.Row(dbc.Col(self.result_card))
            ],className='p-3')

        self.load_callbacks()
        return self.full_layout
    def load_configs(self,default_config='default.csv'):
        configs={x:os.path.join(CONFIG_PATH,x) for x in os.listdir(CONFIG_PATH) if x.endswith('.csv')}
        default_value=configs[default_config] if default_config in configs else None
        return default_config,default_value,[{'label': label, 'value': path} for label,path in configs.items()],configs
    def load_callbacks(self):

        @self([self.config_dropdown.value,self.config_dropdown.options],self.main_app.tabs.active_tab,self.config_dropdown.value)
        def change_configs(active_tab,cur_value):
            if active_tab=='Amplitudes':
                default_config,default_value,config_options,configs=self.load_configs()
                if cur_value not in list(configs.values()):
                    cur_value=default_value
                return cur_value,config_options

        #Load new config into regions table
        @self([self.regions.columns,self.regions.data,self.regions.selected_rows,self.download_config.label,self.config_save_input.value,self.config_save_modal.header],self.config_dropdown.value)
        def change_config(config_path):
            if isinstance(config_path,str) and os.path.isfile(config_path):
                filename=os.path.split(config_path)[1]
                df=pd.read_csv(config_path)
                self.download_config.data=df
                self.download_config.filename=filename
                return [{"name": i, "id": i} for i in df.columns],df.to_dict('records'),list(range(df.shape[0])),f'Download {filename}',filename, f'Save {filename}'

        #Load cur_regions
        @self(self.combiner.children,[self.regions.columns,self.regions.data,self.regions.selected_rows])
        def load_cur_regions(regions_cols,regions_data,selected_rows):
            self.cur_regions=pd.DataFrame.from_records([x for i,x in enumerate(regions_data) if i in selected_rows],columns=[x['name'] for x in regions_cols])
            return 'Config Ready'

        #Save config
        @self([self.config_dropdown.options,self.config_dropdown.value],self.config_save_modal.action.n_clicks,[self.config_save_input.value,self.config_dropdown.options,self.config_dropdown.value,self.regions.columns,self.regions.data])
        def save_config_as(action_clicks,name,options,values,regions_cols,regions_data):
            if name.endswith('.csv') and self.cur_regions is not None:
                path=os.path.join(CONFIG_PATH,name)
                df=pd.DataFrame.from_records(regions_data,columns=[x['name'] for x in regions_cols])
                df.to_csv(path,index=False)
                new_options=[{'label': name, 'value': path}]
                new_values=[path]
                return components.add_to_dropdown(new_options,new_values,options,values,multi=False,keep_old_uploads=True)

        #load Spectra
        self.spectra=[]
        @self(self.combiner.children,self.selected_spectra.value)
        def load_spectra(paths):
            for data in self.spectra:
                data['show']=data['path'] in paths
            for path in paths:
                if os.path.isfile(path):
                    if any([path==x['path'] for x in self.spectra]):
                        continue
                    filename=os.path.split(path)[1]
                    spectra,wl,spectra_nb,baseline=spectra_and_baseline(path)
                    self.spectra.append({'name':filename,
                                        'path':path,
                                        'spectra':spectra,
                                        'wl':wl,
                                        'spectra_nb':spectra_nb,
                                        'baseline':baseline,
                                        'show':True})
            return 'Spectra Ready'


        #RESULTS

        #Figure
        @self([self.graph.figure],[self.combiner.children,self.background.value])
        def display_fig(info,remove_background):
            fig=go.Figure(layout={'height':700,'xaxis_title':'Wavelength (µm)','yaxis_title':'Intensity'})

            for data in self.spectra:
                if data['show']:
                    y=data['spectra'] if len(remove_background)==0 else data['spectra_nb']
                    fig.add_trace(go.Scatter(x=data['wl'], y=y, name=data['name']))

            fig.layout.update(
            shapes=[dict(type="rect",xref="x",yref="paper",x0=float(row['Lower WL']),y0=0,x1=float(row['Upper WL']),y1=1,
                    fillcolor="LightSalmon",line=dict(color="Red",width=1,),opacity=0.5,layer="below") for i,row in self.cur_regions.iterrows()])
            fig.layout.update(showlegend=True)
            return fig

        #Table
        @self([self.results_table.columns,self.results_table.data],[self.combiner.children])
        def display_table(info):
            res_table={'Name':[],'wl start':[],'wl end':[],
                       'Min peak':[],'Peak':[],'Max peak':[],
                       'Min background':[],'Background':[],'Max background':[],
                       'Min S/B':[],'S/B':[],'Max S/B':[],'Pass':[]}

            for data in self.spectra:
               if data['show']:
                   for i,row in self.cur_regions.iterrows():
                       row_val=lambda name,val_if_nan: val_if_nan if row[name] is None or np.isnan(row[name]) else row[name]
                       wl_s=float(row['Lower WL']);wl_e=float(row['Upper WL'])
                       pos=np.logical_and(data['wl']>=wl_s,data['wl']<=wl_e)
                       wl_max=np.argmax(data['spectra'][pos])
                       peak=data['spectra'][pos][wl_max];peak_nb=data['spectra_nb'][pos][wl_max]
                       cur_baseline=peak-peak_nb #baseline[pos][wl_pos]
                       stb=peak/cur_baseline
                       res_table['Name'].append(data['name'])
                       res_table['wl start'].append(wl_s)
                       res_table['wl end'].append(wl_e)
                       res_table['Min peak'].append(row['Min Peak'])
                       res_table['Peak'].append(peak)
                       res_table['Max peak'].append(row['Max Peak'])
                       res_table['Min background'].append(row['Min Background'])
                       res_table['Background'].append(cur_baseline)
                       res_table['Max background'].append(row['Max Background'])
                       res_table['Min S/B'].append(row['Min S/B'])
                       res_table['S/B'].append(stb)
                       res_table['Max S/B'].append(row['Max S/B'])
                       res_table['Pass'].append('Yes' if row_val('Min Peak',-np.inf)<=peak<=row_val('Max Peak',np.inf) and
                                                         row_val('Min Background',-np.inf)<=cur_baseline<=row_val('Max Background',np.inf) and
                                                         row_val('Min S/B',-np.inf)<=stb<=row_val('Max S/B',np.inf) else 'No')

            df=pd.DataFrame(res_table).round(1)
            self.results_table_download.data=df
            return [{"name": i, "id": i} for i in df.columns],df.to_dict('records')



import subprocess, sys, os, hashlib, re
from ipywidgets import widgets
from datetime import datetime
import base64
from io import BytesIO
from IPython.utils.io import capture_output
from IPython.display import display

def get_package(name,install=None):
    try:
        pkg=__import__(name)
    except:
        if install is None:
            install=name
        subprocess.check_call([sys.executable, "-m", "pip", "install", "--trusted-host", "pypi.org", "--trusted-host", "files.pythonhosted.org", install])
        pkg=__import__(name)
    return pkg

get_filename=lambda path: os.path.splitext(os.path.split(path)[-1])[0]

def get_output_path(notebook_path,inputs,folder='Reports',extension='docx'):
    notebook_name=get_filename(notebook_path)
    input_text=''
    if isinstance(inputs,dict):
        process=lambda x: re.sub('\W+',' ', os.path.dirname(x) if os.path.isdir(x) else get_filename(x))[:10]
        input_data=[process(x) for x in list(inputs.values())[:2] if isinstance(x,str)]
        input_text='_'+'_'.join(input_data)
    report_name=f'{notebook_name}{input_text}_{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}.{extension}'
    return os.path.join(folder,report_name)


def get_hash(context):
    return hashlib.sha224(str(context).encode()).hexdigest()

def dict_to_environ(inputs):
    for key,val in inputs.items():
        if isinstance(key,str) and isinstance(val,str):
            os.environ[key]=val

def capture_code(code,scope=None):
    if scope is None:
        scope={}

    #import pdb;pdb.set_trace()
    print('---------------------------CODE----------------------------------------')
    print(code)
    with capture_output() as capture:

        exec(compile(code,"na","exec"),scope,scope)

    return capture,scope

def run_code_in_shell(code):
    from IPython.utils.io import capture_output
    run_code=("python -c 'from IPython.core.interactiveshell import InteractiveShell\n"+
          "shell=InteractiveShell().instance()\n"+
          "from IPython.utils.io import capture_output\n"+
          "co=capture_output()\ncapture=co.__enter__()\n"+
          code+"\nco.__exit__(1,2,3)\n"+
          "for output in capture.outputs:\n"+
          "\tif \"text/html\" in output.data:\n"+
          "\t\tprint(output.data[\"text/html\"])'")
    proc = subprocess.Popen(run_code, shell=True,
                         stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    return proc.communicate()[0].decode('utf-8')

    #Example ------------------------------
    import nbformat
    path='/home/jovyan/shared/BTR/Standard Report/Types/Snapshot.ipynb'
    notebook=nbformat.read(path, as_version=4)
    for i,cell in enumerate(notebook['cells']):
        if cell['cell_type']=='code':
            code=''.join([line for line in cell['source'].split('\n') if not line.startswith('%')])
            print(run_code(code))

def outputs_to_html(outputs,types=['text/html','application/vnd.jupyter.widget-view+json']):
    res=''
    removes=[r'<a.*?>|<\/a.*?>','<title>.+</title>']
    if not isinstance(outputs,list) and hasattr(outputs,'outputs'):
        outputs=outputs.outputs

    for output in outputs:
        if types is None or any([x in output.data for x in types]):
            if 'text/html' in output.data:
                res+=output.data['text/html']
            elif 'application/vnd.jupyter.widget-view+json' in output.data and 'model_id' in output.data['application/vnd.jupyter.widget-view+json']:
                model_id=output.data['application/vnd.jupyter.widget-view+json']['model_id']
                if model_id in widgets.Widget.widgets:
                    widget=widgets.Widget.widgets[model_id]
                    if hasattr(widget,'figure'):
                        fig=widget.figure
                        fig.tight_layout()
                        tmpfile = BytesIO()
                        fig.savefig(tmpfile, format='png')
                        encoded = base64.b64encode(tmpfile.getvalue()).decode('utf-8')
                        res+='<p><img src=\'data:image/png;base64,{}\'></p>'.format(encoded)
            elif 'image/png' in output.data:
                encoded = base64.b64encode(output.data['image/png']).decode('utf-8')
                res+='<p><img src=\'data:image/png;base64,{}\'></p>'.format(encoded)
            else:
                res+=output.data['text/plain']

    for pattern in removes:
        res=re.sub(r'<a.*?>|<\/a.*?>','',res,re.I | re.M)

    return res

def outputs_to_output(outputs,ipyoutput,types=['text/html','application/vnd.jupyter.widget-view+json']):
    if not isinstance(outputs,list) and hasattr(outputs,'outputs'):
        outputs=outputs.outputs
    with ipyoutput:
        for output in outputs:
            if types is None or not hasattr(output,'data') or any([x in output.data for x in types]):
                display(output)


def get_outputs_figure_widgets(outputs):
    res=[]
    if not isinstance(outputs,list) and hasattr(outputs,'outputs'):
        outputs=outputs.outputs
    for output in outputs:
        if 'application/vnd.jupyter.widget-view+json' in output.data and 'model_id' in output.data['application/vnd.jupyter.widget-view+json']:
            model_id=output.data['application/vnd.jupyter.widget-view+json']['model_id']
            if model_id in widgets.Widget.widgets:
                canvas=widgets.Widget.widgets[model_id]
                canvas.figure.tight_layout()
                canvas.header_visible = False
                canvas.footer_visible = False
                res.append(canvas)
    return res



from .base import *
from .base_component import Component, Item
from flask import send_file,make_response
import pandas as pd
import threading
import weakref

def serve_df(name,df):
    resp = make_response(df.to_csv(index=False))
    resp.headers["Content-Disposition"] = f"attachment; filename={name}"
    resp.headers["Content-Type"] = "text/csv"
    return resp

class Download(Component):
    registered=False
    def __init__(self,filename=None,data=None,style='button',small=False):
        self.filename=filename
        self.data=data
        instances=type(self).get_instances()
        instances[self.id]=self#weakref.proxy(self)

        children=''
        if filename is not None:
            children=f'Download {filename}'
        self.layout = html.A(children=children,target='_blank',href=f'fetch/{self.id}')

        if style=='button':
            self.layout.obj.className="btn btn-primary"+(' btn-sm' if small else '')

        self.properties={'url':self.layout.href,
                        'label':self.layout.children}

    @classmethod
    def register(cls,app,prefix='/'):
        @app.server.route(prefix+"fetch/<string:download_id>")
        def download_table(download_id):
            return cls.serve(download_id)
        cls.registered=True
    @classmethod
    def serve(cls,download_id):
        downloads=cls.get_instances()
        if download_id in downloads:
            download=downloads[download_id]
            data=download.data
            filename=download.filename
            if isinstance(data,pd.DataFrame):
                return serve_df(filename,data)
            elif isinstance(data,str) and os.path.isfile(data):
                return send_file(data,attachment_filename=filename)
        return 'Resource not found',404
    @classmethod
    def get_instances(cls):
        if not hasattr(cls, 'instances'):
            cls.instances = {}
        return cls.instances



import os

def data_to_bytes(data):
    if isinstance(data, str):
        if os.path.isfile(data):
            with open(data, 'rb') as f:
                data = f.read()
        else:
            data=data.encode()

    return data



from flask import Flask,make_response,session
import pandas as pd
import numpy as np
import uuid
import dash
import os

import webbrowser

def get_dynamic_values(id_type):
    res={}
    for inputs_list in dash.callback_context.inputs_list:
        if isinstance(inputs_list,list): #dynamic types
            for cur_input in inputs_list:
                if 'value' in cur_input and isinstance(cur_input['id'],dict) and cur_input['id']['type']==id_type:
                    res[cur_input['id']['index']]=cur_input['value']
    return res

def generate_unique_id():
    return uuid.uuid4().hex[:6].upper()

def open_browser():
    webbrowser.open_new('http://127.0.0.1:5000/cc/')

def load_environ_variables():
    with open('.env') as f:
        for line in f:
            if not line.startswith('#') and '=' in line:
                props=line.split('=')
                if len(props)==2 and not props[0] in os.environ:
                    os.environ[props[0]]=props[1]



import dash

def get_dynamic_values(id_type):
    res={}
    for inputs_list in dash.callback_context.inputs_list:
        if isinstance(inputs_list,list): #dynamic types
            for cur_input in inputs_list:
                if 'value' in cur_input and isinstance(cur_input['id'],dict) and cur_input['id']['type']==id_type:
                    res[cur_input['id']['index']]=cur_input['value']
    return res



from dash.dependencies import Input, Output, State, MATCH, ALL
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import plotly.graph_objs as go
import dash_table
import pandas as pd



def baseline_als(y, lam=100, p=0.001, n_iter=10, max_mse=0,L=None):
    """
    Estimate baseline of spectrum using the Assymetric Least Squares [1] algorithm.

    Args:
        y:          Input spectrum.
        lam:        (Optional) Smoothing parameter.
        p:          (Optional) Assymetry parameter.
        n_iter:     (Optional) Maximum number of fitting iterations.
        max_mse:    (Optional) Maximum MSE change between fitting iterations that
                    will terminate the fitting process.

    Returns:
        z:          Estimated baseline.

    References:
        [1]     Sung-June Baek, et. al. Baseline correction using asymmetrically
                reweighted penalized least squares smoothing (Analyst, 2015, 140, 250)
    """

    import numpy as np
    from scipy import sparse
    from scipy.sparse.linalg import spsolve

    L = len(y)
    D = sparse.csc_matrix(np.diff(np.eye(L), 2))
    w = np.ones(L)
    for i in range(0, n_iter):
        W = sparse.spdiags(w, 0, L, L)
        Z = W + lam * D.dot(D.transpose())
        if i > 0:
            z_prev = z
        z = spsolve(Z, w * y)
        if i > 0:
            mse = np.sqrt(np.sum(z - z_prev) ** 2) / len(z)
            if mse < max_mse:
                break
        w = p * (y > z) + (1 - p) * (y < z)

    return y-z,z



import base64
from io import BytesIO
import re

def create_wrap(debug_view):
    def wrap(control=None,observe=True,**wrapkwargs):
        def decor(fcn):
            @debug_view.capture(clear_output=True)
            def rtn(*args,**kwargs):
                print(f'Function: {fcn.__name__} with args {args} and kwargs {kwargs}')
                if control and len(args)==0 and len(kwargs)==0:
                    args=[control.value]
                else:
                    val=lambda x: x['new'] if isinstance(x,dict) and 'new' in x else x.value if hasattr(x,'value') else x
                    args=[val(x) for x in args]
                    kwargs={key:val(x) for key,x in kwargs.items()}
                return fcn(*args,**kwargs,**wrapkwargs)

            if control and observe:
                control.observe(rtn,names=['value'])
            return rtn
        return decor
    return wrap

def get_output_figure_widgets(output):
    res=[]
    for elem in output.outputs:
        if elem['output_type']=='display_data':
            if 'application/vnd.jupyter.widget-view+json' in elem['data'] and 'model_id' in elem['data']['application/vnd.jupyter.widget-view+json']:
                res.append(output.widgets[elem['data']['application/vnd.jupyter.widget-view+json']['model_id']])
    return res

def get_output_html(output):
    res=''
    removes=[r'<a.*?>|<\/a.*?>','<title>.+</title>']
    for elem in output.outputs:
        if elem['output_type']=='display_data':
            html=None
            if 'text/html' in elem['data']:
                res+=elem['data']['text/html']
            elif 'application/vnd.jupyter.widget-view+json' in elem['data'] and 'model_id' in elem['data']['application/vnd.jupyter.widget-view+json']:
                fig=output.widgets[elem['data']['application/vnd.jupyter.widget-view+json']['model_id']].figure
                tmpfile = BytesIO()
                fig.savefig(tmpfile, format='png')
                encoded = base64.b64encode(tmpfile.getvalue()).decode('utf-8')

                res+='<img src=\'data:image/png;base64,{}\'>'.format(encoded)
            elif 'text/plain' in elem['data']:
                res+=elem['data']['text/plain']

    for pattern in removes:
        res=re.sub(r'<a.*?>|<\/a.*?>','',res,re.I | re.M)

    return res



import pickle
import os 
import hashlib
from datetime import datetime
import pandas as pd

def cache_output(path=None,max_cache_seconds=259200):
    if path is None:
        path = os.path.join(os.path.dirname(os.path.realpath(__file__)),'cache')
        if not os.path.isdir(path):
            os.mkdir(path)
            
    def wrap(fcn):
        def rtn(*args,**kwargs):
            # Remove old cached files
            max_time_path=os.path.join(path,'max times.csv')
            if os.path.isfile(max_time_path):
                df=pd.read_csv(max_time_path)
                remove_pos=df.remove_timestamp<=datetime.now().timestamp()
                if df[remove_pos].shape[0]>0:
                    for index,row in df[remove_pos].iterrows():
                        file_path=os.path.join(path,row.file)
                        if os.path.isfile(file_path):
                            os.remove(file_path)
                    df[~remove_pos].to_csv(max_time_path,index=False)
            else:
                with open(max_time_path, 'w') as f:
                    f.write('file,remove_timestamp\n')
            
            #Check if hashed file already exists
            identifier_text=f'{fcn.__name__}-{type(fcn).__mro__}-{args}-{kwargs}'
            hash = hashlib.md5(identifier_text.encode()).hexdigest()
            save_file=f'{hash}.pkl'
            full_path=os.path.join(path,save_file)
            if os.path.isfile(full_path):
                with open(full_path, 'rb') as f:
                    return pickle.load(f)
                
            #Get and writer results to hash file
            res=fcn(*args,**kwargs)
            with open(full_path, 'wb') as f:
                pickle.dump(res, f, pickle.HIGHEST_PROTOCOL)
            
            #Append removal time for hashed file
            if max_cache_seconds is not None:
                with open(max_time_path, 'a') as f:
                    f.write(f'{save_file},{datetime.now().timestamp()+max_cache_seconds}\n')
            return res
        return rtn
    return wrap


from .. import components
from ..components.base import *
from ..config import *
from ..utils import spectra_and_baseline
import pandas as pd
import numpy as np
from datetime import datetime

class EditFiles(components.Component):
    def __init__(self,name):
        self.dropdown=components.Dropdown(multi=True,options=None,update_value_on_change=True,keep_old_values=True,value=None)
        self.uploader=components.Uploader(type='area',multi=True,allowed_extensions=SPECTRA_EXTENSIONS,upload_to=SPECTRA_PATH,dropdown=self.dropdown,keep_old_uploads=True)
        self.layout=html.Div([
                        html.Label(f'{name}:'),
                        self.uploader,
                        dcc.Loading(self.dropdown)
                    ],className='border')

class Edit(components.Component):
    def __init__(self,overview):
        self.alert=components.Alert()

        self.new_unit_name=dcc.Input(value='',placeholder='Enter unit name...',style={'width':'600px'},className='ml-2')
        self.new_unit_loading=html.Div()
        self.add_unit=html.A('Add Unit',href='#',className="btn btn-primary mr-3",style={'width':'100px'})
        self.new_unit_clear=html.A('Clear',href='#',className="btn btn-secondary mr-3")

        self.coupons=EditFiles('Coupons')
        self.particles=EditFiles('Particles')

        self.results_table_download=components.Download(filename='Results.csv')
        self.delete_units=html.Div('')
        self.delete_modal=components.Modal(open_label='Delete Rows',open_color='danger',header='Delete Rows',body=html.Div(['Are you sure you want to delete the following units:',self.delete_units]),action_label='Delete',action_color='danger',show_cancel=True,small=True)
        self.download_results=components.Download(filename='Results.csv',data=overview.metrics,small=True).className('ml-3 mr-3')
        self.results_table=dash_table.DataTable(columns=[{"name": i, "id": i} for i in overview.metrics.columns],data=overview.metrics.to_dict('records'),
                                                row_selectable='multi',selected_rows=[],editable=False,css=[{'selector': '.row', 'rule': 'margin: 0'}])

        self.layout=html.Div([
            self.alert,
            dbc.Row(
                dbc.Card([
                    dbc.CardHeader(dbc.Row([
                        dbc.Col(['New Unit:',self.new_unit_name]),
                        dbc.Col(dcc.Loading(self.new_unit_loading))
                    ])),
                    dbc.CardBody([
                        dbc.Row([dbc.Col(self.coupons),dbc.Col(self.particles)]),
                        dbc.Row([dbc.Col(self.new_unit_clear),self.add_unit],justify='between',className='pt-3')
                    ])
                ,],className='w-100 m-4')),
            dbc.Row(
                dbc.Card([
                    dbc.CardHeader(dbc.Row(['Results:',self.delete_modal.open,self.download_results])),
                    dbc.CardBody([
                        #dbc.Row([self.delete_modal.open,self.download_results],justify="start",className='pb-2'),
                        dbc.Row(dcc.Loading(self.results_table)),
                    ])
                ,],className='w-100 m-4')),
            self.delete_modal
        ])

        #Clear uploads
        self.clear_clicks=0
        @self([self.coupons.dropdown.value,self.coupons.dropdown.options,self.particles.dropdown.value,self.particles.dropdown.options],[self.new_unit_clear.n_clicks,overview.info],[self.coupons.dropdown.value,self.particles.dropdown.value])
        def clear_uploads(clicks,info,coupon_paths,particle_paths):
            clear_now=self.clear_clicks!=clicks or info.startswith('Added')
            self.clear_clicks=clicks
            if clear_now:
                for path in coupon_paths+particle_paths:
                    if os.path.isfile(path):
                        os.remove(path)
                return [],[],[],[]

        #Add rows to delete modal
        @self(self.delete_units.children,[self.results_table.columns,self.results_table.data,self.results_table.selected_rows])
        def change_row_selection(cols,data,selected_rows):
            df=pd.DataFrame.from_records([x for i,x in enumerate(data) if i in selected_rows],columns=[x['name'] for x in cols])
            return ', '.join(list(df.SN.unique()))

        #Delete rows
        @self([overview.info,self.alert.alerts],self.delete_modal.action.n_clicks,[self.results_table.columns,self.results_table.data,self.results_table.selected_rows,self.alert.alerts])
        def delete_rows(clicks,cols,data,selected_rows,alerts):
            if alerts is None: alerts=[]
            df=pd.DataFrame.from_records([x for i,x in enumerate(data) if i not in selected_rows],columns=[x['name'] for x in cols])
            n_rows=overview.metrics.shape[0]-df.shape[0]
            overview.metrics=df
            alerts.append(self.alert.create(f'Successfully deleted {n_rows} rows','success'))
            return f'Deleted {n_rows} rows',alerts

        #get Results
        @self([overview.info,self.alert.alerts,self.new_unit_loading.children],self.add_unit.n_clicks,[self.new_unit_name.value,self.coupons.dropdown.value,self.particles.dropdown.value,self.alert.alerts])
        def get_results(clicks,name,coupons,particles,alerts):
            if alerts is None: alerts=[]
            if len(name)<4:
                alerts.append(self.alert.create('Please enter a longer name for the new unit','info'))
                return 'missing name',alerts,''
            if ',' in name:
                alerts.append(self.alert.create('You cannot use commas in the unit name','info'))
                return 'commas',alerts,''
            def path_res(path):
                spectra,wl,spectra_nb,baseline=spectra_and_baseline(path)
                wl_s=315;wl_e=316
                pos=np.logical_and(wl>=wl_s,wl<=wl_e)
                wl_max=np.argmax(spectra[pos])
                peak=spectra[pos][wl_max];peak_nb=spectra_nb[pos][wl_max]
                background=peak-peak_nb
                return peak,peak_nb,background,peak_nb/background
            def mean_res(paths):
                peaks=[];peak_nbs=[];backgrounds=[];sbrs=[]
                for i,path in enumerate(paths):
                    peak,peak_nb,background,sbr=path_res(path)
                    peaks.append(peak);peak_nbs.append(peak_nb);backgrounds.append(background);sbrs.append(sbr)
                return np.mean(peaks),np.mean(peak_nbs),np.mean(backgrounds),np.mean(sbrs)

            cpn_res=mean_res(coupons)
            ptl_res=mean_res(particles)

            now=datetime.now().replace(microsecond=0)
            df=pd.DataFrame({'SN':[name],'Date':[now],
                            'Coupon':[cpn_res[1]],'Coupon Background':[cpn_res[2]],'Coupon SBR':[cpn_res[3]],
                            'Particle':[ptl_res[1]],'Particle Background':[ptl_res[2]],'Particle SBR':[ptl_res[3]]}).round(2)

            overview.metrics=pd.concat([overview.metrics,df])
            alerts.append(self.alert.create(f'{name} added successfully','success'))
            return f'Added {name}',alerts,''

        #Change table and save metrics
        @self([self.results_table.columns,self.results_table.data,self.results_table.selected_rows],overview.info)
        def metrics_changed(info):
            overview.metrics.to_csv(METRICS_PATH,index=False)
            self.download_results.data=overview.metrics
            return [{"name": i, "id": i} for i in overview.metrics.columns],overview.metrics.to_dict('records'),[]



import tkinter as tk

from .utils import Component

class FrameCol(object):


class Frame(tk.Frame,Component):
    _cols=None
    def __init__(self,cols=1,rows=1,*args,**kwargs):
        """ Initialize a frame with a grid layout
            The grid is defined by the inputs rows and cols
            Use indices to get a cell frame as [column][row]
            E.g.
                frame = Frame(2,2)
                cell_12 = frame[1][2]
            Inputs:
                rows,cols - int or [dict]
                    if integer - defines the number of equal spaced default sytle rows/cols
                    if a single/list of dictionaries - defines the properties of each row or col
                    E.g. rows=3, cols=3
                            or
                        rows=[{'weight':1, 'minsize':75},{'weight':1, 'minsize':100}]
        """
        super().__init__(*args,**kwargs)
        if isinstance(cols,int): cols=[{}]*cols
        if isinstance(rows,int): rows=[{}]*rows
        if type(rows) not in [list,tuple]:rows=[rows]
        if type(cols) not in [list,tuple]:cols=[cols]
        for i_row,row in enumerate(rows):
            self.columnconfigure(i_row, **col)
            for i_col,col in enumerate(cols):
                if i_row==0:
                    self.columnconfigure(i_col, **col)

    def get_border_effect(self,name):
        border_effects = {
            "flat": tk.FLAT,
            "sunken": tk.SUNKEN,
            "raised": tk.RAISED,
            "groove": tk.GROOVE,
            "ridge": tk.RIDGE,
        }
        return border_effects[name]

    def __getitem__(object,i_col):
        if self._cols is None



import pandas as pd
from mongoengine import *
import json

from mongoengine.queryset.queryset import QuerySet
def to_df(self,exclude=None):
    if exclude is None: exclude=['_id']
    return pd.DataFrame.from_records(json.loads(self.to_json()),exclude=exclude)
QuerySet.to_df=to_df

def Document(DynamicDocument):
    pass



import uuid
import re
import os
import logging
import sys
import psutil

from . import functions

def kill_proc_tree(pid, including_parent=True):
    parent = psutil.Process(pid)
    children = parent.children(recursive=True)
    for child in children:
        child.kill()
    gone, still_alive = psutil.wait_procs(children, timeout=5)
    if including_parent:
        parent.kill()
        parent.wait(5)

def get_cwd():
    return os.path.abspath(os.getcwd())

def get_base_path():
    return os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

#file_dir = os.path.dirname(os.path.realpath(__file__))
def append_fcns(fcn,*fcns):
    def rtn(*args,**kwargs):
        res=[x(*args,**kwargs) for x in ([fcn]+fcns)]
        if any([x is False for x in res]):
            return False
    return rtn

def create_logger(name,level='INFO',add_stdout_handler=True):
    logger = logging.getLogger(name)
    ch = logging.StreamHandler(sys.stdout)
    level=getattr(logging,level)
    logger.setLevel(level)

    if add_stdout_handler:
        ch.setLevel(level)
        formatter = logging.Formatter('[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    return logger

def create_uuid(hex=True):
    res=uuid.uuid4()
    return res.hex if hex else str(res)

def is_list(obj):
    return type(obj) in [list,tuple]

def iterable(l):
    try:
        iter(l)
        return True
    except:
        return False

def natural_sort(l,by=None):
    convert = lambda text: int(text) if text.isdigit() else text.lower()

    pos=None
    if iterable(by) and len(by)==len(l):
        l=zip(l,by)
        pos=1
    alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key if pos is None else key[pos]) ]
    res=sorted(l, key = alphanum_key)
    if pos is not None: res=list(zip(*res))[0]
    return res

def combine_decorators(decorators):
    def rtn_func(func):
        func=func
        for decorator in decorators[::-1]:
            func=decorator(func)
        return func
    return rtn_func

comparitors=["exact","in","gte","lte","lt","gt","contains","icontains","istartswith","iendswith","startswith","endswith","year","month","day","hour","minute","second","isnull"]
def compare_input_to_value(input,value,comparison=None):
    if comparison is None or comparison=="exact":
        return input==value
    elif comparison=="in":
        return input in value
    elif comparison=="gte":
        return input >= value
    elif comparison=="lte":
        return input <= value
    elif comparison=="lt":
        return input < value
    elif comparison=="gt":
        return input > value
    elif comparison=="contains":
        return value in input
    elif comparison=="icontains":
        return value.lower() in input.lower()
    elif comparison=="startswith":
        return input.startswith(value)
    elif comparison=="endswith":
        return input.endswith(value)
    elif comparison=="istartswith":
        return input.lower().startswith(value.lower())
    elif comparison=="iendswith":
        return input.lower().endswith(value.lower())
    elif comparison in ["year","month","day","hour","minute","second"]:
        return getattr(input,comparison)==value
    elif comparison=="isnull":
        return input is None
    else:
        raise Exception("Comparison not allowed with: %s"%comparison)

def get_field(obj,prop,setting_field=False,method_kwargs=None):
    props=prop.replace(".","__").split("__")
    field=obj
    last_field=obj;last_type=None
    for i_prop,prop in enumerate(props):
        last_field=field
        if hasattr(field,prop):
            prop_val=getattr(field,prop)
            last_type="attr"

            if inspect.ismethod(prop_val):
                available_inputs=["self"]
                if method_kwargs is None:
                    method_kwargs={}
                else:
                    available_inputs.extend(list(method_kwargs))
                input_names=functions.need_more_inputs(prop_val,available_inputs=available_inputs,return_missing=True)
                if len(input_names)>0:
                    return "Missing inputs: %s"%", ".join(input_names)
                else:
                    field=prop_val(**method_kwargs)
            else:
                field=prop_val
        elif prop in comparitors:
            last_type="comparison"
            field=lambda compare_to: compare_input_to_value(field,compare_to,prop)
        elif isinstance(field, dict):
            last_type="dict"
            assert prop in field, f"Key {prop} is not in field"
            field=field[prop]
        elif isinstance(field, list): #prop must include "name#type" or "type" if adding a new index
            last_type="list"
            if is_integer(prop) and has_index(field,prop):
                field = field[int(prop)]
            elif prop=='min':
                field=min(field)
            elif prop=='max':
                field=max(field)
            elif prop=='mean':
                field=np.mean(field)
            elif ':' in prop:
                field=field[slice(*[int(x) for x in prop.split(':')])]
            else:
                raise Exception(f'{prop} not found in list or one of [min, max, mean]')
        else:
            field=None

    if setting_field:
        return last_field,prop,last_type,field
    else:
        return field



from ..utils import *
from ..notebook import get_notebook
import pandas as pd
import os, json

csv_path='reports.csv'

def get_report_hash(notebook_path,inputs):
    nb=[x['source'] for x in get_notebook(notebook_path).cells if 'source' in x]
    context=f'{str(nb)}-{str(inputs)}'
    return get_hash(context)

def get_previous_report(notebook_path,inputs,csv_path=csv_path):
    hash=get_report_hash(notebook_path,inputs)
    if os.path.isfile(csv_path):
        report_df=pd.read_csv(csv_path)
        report_df=report_df[report_df.hash==hash]
        if report_df.shape[0]>0:
            paths=report_df.path.iloc[-1].split('#')
            if all([os.path.isfile(x) for x in paths]):
                return paths
    return None

def get_previous_inputs(notebook=None,inputs=None,csv_path=csv_path):
    res={}
    if inputs is not None:
        if not isinstance(inputs,list):
            inputs=[inputs]
        res={x:[] for x in inputs}

    if os.path.isfile(csv_path):
        df=pd.read_csv(csv_path)
        if notebook is not None:
            df=df[df.notebook==notebook]
        for i,row in df.iterrows():
            vals=json.loads(row.inputs)
            for key,val in vals.items():
                if inputs is None and key not in res:
                    res[key]=[]
                if key in res and val not in res[key]:
                    res[key].append(val)
    if len(res)==1:
        res=res[list(res.keys())[0]]
    return res

def record_report(notebook_path,inputs,paths,csv_path=csv_path):
    hash=get_report_hash(notebook_path,inputs)
    paths='#'.join(paths)
    df=pd.DataFrame({'hash':[hash],'path':[paths],'inputs':json.dumps(inputs),'notebook':notebook_path})
    if os.path.isfile(csv_path):
        report_df=pd.read_csv(csv_path)
        report_df=report_df[report_df.hash!=hash]
        df=report_df.append(df, ignore_index=True)
    df.to_csv(csv_path,index=False)



from .base import *
from .base_component import Component, Item

class Tabs(Component):
    def __init__(self,contents,active_tab=None):
        labels=list(contents.keys())
        if active_tab is None or active_tab not in labels:
            active_tab=labels[0]
        self.tabs=dbc.Tabs([dbc.Tab(label=label, tab_id=label, style={'cursor': 'pointer'}) for label in labels],
                card=True,active_tab=active_tab)

        to_item = lambda label,content: content if isinstance(content,Item) else Item(self,label,content)
        item_contents=[to_item(f'content-{label}',html.Div(content,style={'display': 'block' if label==labels[0] else 'none'})) for label,content in contents.items()]
        self.layout = html.Div(
                        dbc.Card([
                            dbc.CardHeader(self.tabs),
                            dbc.CardBody(item_contents)
                        ,])
                    ,className='pt-3')

        self.properties={'active_tab':self.tabs.active_tab}

        @self([x.style for x in item_contents],self.active_tab)
        def change_content(active_tab):
            return [{'display': 'block' if label==active_tab else 'none'} for label in labels]



from .base import *
from .base_component import Component

def add_to_dropdown(new_options,new_values,options,values,multi=True,keep_old_uploads=True):
    if keep_old_uploads:
        options=options+[x for x in new_options if x not in options]
    else:
        options=new_options

    if keep_old_uploads and multi:
        values=values+[x for x in new_values if x not in values]
    elif multi:
        values=new_values
    elif len(new_values)>0:
        values=new_values[0]

    return options,values

class Dropdown(Component):
    def __init__(self,multi=False,label=None,options=None,update_value_on_change=True,keep_old_values=True,value=None):

        if options is None: options=[]
        options = [{'label': x, 'value': x} if isinstance(x,str) else x for x in options]
        if value is None: value=[]

        self.dropdown=dcc.Dropdown(options=options,multi=multi,value=value)

        children=[html.Label([f'{label}: '])] if label else []
        children.append(self.dropdown)

        self.layout=html.Div(children)

        self.properties={'value':self.dropdown.value,
                        'options':self.dropdown.options,
                        'loading_state':self.dropdown.loading_state}



import sqlalchemy as sql
import sqlalchemy.orm as orm
from copy import copy
from functools import partial
import os
import pandas as pd

from .db import *

def update_class(obj,new_class):
    for key,val in dict(new_class.__dict__).items():
        if key not in ['__module__','__dict__','__weakref__','__doc__']:
            setattr(obj,key,val)

def get_object_with_instance(obj,class_name=None):
    return [val for key,val in globals().items() if (hasattr(val,'__dict__') and obj in val.__dict__.values()) and (class_name is None or val.__class__.__name__==class_name)]

class tableMetaData:
    def __init__(self,tables):
        if type(tables) not in [list,tuple]: tables=[tables]
        self.tables=tables
    @property
    def columns(self):
        return [x for table in self.tables for x in list(table.columns)]
    @property
    def fields(self):
        props=['name','type','unique','default','nullable','comment','foreign_keys']
        df=pd.concat([pd.DataFrame([{x:getattr(col,x) for x in props} for col in table.columns]).assign(table=table.name).set_index(['name']) for table in self.tables])
        return df
    @property
    def field_names(self):
        return [col.name for table in self.tables for col in table.columns]

    @property
    def foreign_keys(self):
        res={}
        for col in self.columns:
            for fk in col.foreign_keys:
                if col.name not in res:
                    res[col.name]=[]
                res[col.name].append(fk)
        return res
    @property
    def foreign_tables(self):
        return [fk.column.table for col_name,fks in self.foreign_keys.items() for fk in fks]

    def related(self,related_type='any'): #related_type = ['any','upstream','downstream']
        related_type=related_type.lower()
        assert related_type in ['any','upstream','downstream'], "related_type must be one of: ['any','upstream','downstream']"
        res=[]
        if related_type in ['any','downstream']:
            for col_name,fks in self.foreign_keys.items():
                for fk in fks:
                    res.append({'column':col_name,'related_column':fk.column.name,'table':fk.column.table,'type':'downstream','fk':fk})
        if related_type in ['any','upstream']:
            for table in self.tables[0].tables:
                for col_name,fks in table._meta.foreign_keys.items():
                    for fk in fks:
                        if fk.column.table in self.tables:
                            res.append({'column':fk.parent.name,'related_column':col_name,'table':fk.parent.table,'type':'upstream','fk':fk})
        return pd.DataFrame(res)

def get_query_for_foreign_key(fk,objs,objs_for_parent=True):
    parent_table=orm.aliased(fk.parent.table)
    column_table=orm.aliased(fk.column.table)
    if objs_for_parent:
        rel_table=column_table
        filter_col=column_table.columns[fk.column.name]; objs_col=parent_table.columns[fk.parent.name]
        pks=list(parent_table.primary_key)
        filters=filter_for_rows(objs,pks)
    else:
        filters=[]
        rel_table=parent_table;
        filter_col=parent_table.columns[fk.parent.name]; objs_col=column_table.columns[fk.column.name]
    filters+=filter_for_rows(objs,[filter_col],[objs_col])
    return _query(rel_table,session=fk.parent.table.sess).filter(*filters)

class queryResult(object):
    def __init__(self,result,tables):
        self.tables=tables
        self.sess=tables[0].sess
        self._fields=result._fields
        self.__dict__.update({field:getattr(result,field) for field in self._fields})
        self._meta=tableMetaData(tables)
        self.all_tables=tables[0].tables
    def __str__(self):
        return str({field:getattr(self,field) for field in self._fields})
    def __repr__(self):
        return str(self)
    @property
    def objects(self):
        return _query(self.tables,session=self.sess)
    def search(self,*entities):
        return _query(*entities,session=self.sess)
    def related_queries(self,targets=None,related_type='any'): #related_type = ['any','upstream','downstream']
        related_df=self._meta.related(related_type=related_type)

        if targets is None: targets=list(self.all_tables)
        if type(targets) not in [list,tuple]: targets=[targets]
        targets=self.tables[0].tables.get(targets)

        related_df=related_df[related_df.table.isin(targets)]
        #assert related_df.shape[0]>0, f'No relationship were found for {", ".join([x.name for x in targets])}.'

        res=[]
        for i,row in related_df.iterrows():
            q=get_query_for_foreign_key(row.fk,[self],objs_for_parent=(row.type=='downstream'))
            res.append((row.column,row.table,row.type,q,q.count()))
        return pd.DataFrame(res,columns=['column','table','type','dbquery','count'])

    def relations(self,targets=None,related_type='upstream',only_one=True):
        res=self.related_queries(targets=targets,related_type=related_type)
        items=[x for q in res.dbquery.to_list() for x in q.all()]
        return items[0] if only_one and len(items)>0 else items

    def __getattr__(self, name):
        if name.startswith('get_'):
            fk=self._get_fk(name[4:])
            if fk is not None:
                return self.get_fk_item(fk)
        for table in self.tables:
            if name in table.extra_attribs:
                res=table.extra_attribs[name]
                if callable(res):
                    return res(self)
                else:
                    return res
        raise AttributeError(f'{self.__class__.__name__}.{name} is invalid.')
    @property
    def fks(self):
        if not hasattr(self,'_fks'):
            self._fks={fk.parent.name:fk for table in self.tables for fk in table.foreign_keys}
        return self._fks
    def _get_fk(self,fk_name):
        fks=self.fks
        return fks[fk_name] if fk_name in fks else None
    def get_fk_item(self,fk,limit=None):
        q=get_query_for_foreign_key(fk,[self],objs_for_parent=True)
        if limit is not None: q=q.limit(limit)
        items=q.all()
        return items

class _query(orm.query.Query):
    @property
    def tables(self):
        if not hasattr(self,'_tables'):
            tables=[]
            for entity in self._raw_columns:
                table=entity#.column.table
                tables.append(table.original if isinstance(table,sql.sql.selectable.Alias) else table)
            self._tables=list(set(tables))
        return self._tables
    def __getitem__(self, *args,**kwargs):
        results=super().all().__getitem__(*args,**kwargs)
        return self._add_related_to_results(results)
    def all(self, *args,**kwargs):
        results=super().all(*args,**kwargs)
        return self._add_related_to_results(results)
    @property
    def df(self):
        return query_to_df(self)
    def _add_related_to_results(self,results):
        is_single=False
        if not isinstance(results,list):
            results=[results]
            is_single=True
        new_results=[]
        for result in results:
            new_results.append(queryResult(result,self.tables))
        return new_results[0] if is_single else new_results

    def filter(self,*args,**kwargs):
        filters=convert_props_to_filters(self.tables,*args,**kwargs)
        return super().filter(*filters)

    def _table_with_column(self,col_name,assert_exists=False):
        return table_with_column(self.tables,col_name,assert_exists=assert_exists)

    def related(self,col_name):
        table=self._table_with_column(col_name,assert_exists=True)
        col=getattr(table.columns,col_name)
        fks=list(col.foreign_keys)
        return get_query_for_foreign_key(fks[0],self,objs_for_parent=True) if len(fks)>0 else None

class Column(sql.sql.schema.Column):
    def change_foreign_keys(self):
        for fk in self.foreign_keys:
            fk.parent.table=self.table.tables.get(fk.parent.table)
            fk.column.table=self.table.tables.get(fk.column.table)

class Table(sql.sql.schema.Table):
    def change_foreign_keys(self):
        for fk in self.foreign_keys:
            fk.parent.table=self.tables.get(fk.parent.table)
            fk.column.table=self.tables.get(fk.column.table)
    @property
    def _meta(self):
        return tableMetaData(self)
    def add_result_attribs(self,attribs):
        self.extra_attribs.update(attribs)
    @property
    def objects(self):
        return _query(self,session=self.sess)

class tables(object):
    iter_pos = 0
    table_names=[]
    def __init__(self,DB_instance):
        table_names=DB_instance.table_names
        self.sess=DB_instance.sess
        if table_names is None:
            table_names=get_table_names(DB_instance.meta)
        schema='' if DB_instance.schema is None else f'{DB_instance.schema}.'

        #add tables to self
        for table_name in table_names:
            if len(table_name.split('.'))==1:
                table_name=f'{schema}{table_name}'
            table=DB_instance.meta.tables[table_name]
            self.add_table(table)
    def __iter__(self):
        self.iter_pos = 0
        return self
    def __getitem__(self,name):
        return self.get(name)
    def __next__(self):
        if self.iter_pos < len(self.table_names):
            result=self.get(self.table_names[self.iter_pos])
            self.iter_pos += 1
            return result
        else:
            raise StopIteration

    def __str__(self):
        return ', '.join(self.table_names)
    def add_table(self,table):
        if table.name not in self.__dict__:
            self.table_names.append(table.name)
            table.__class__=Table
            table.tables=self
            table.sess=self.sess
            table.extra_attribs={}
            self.__dict__[table.name]=table
            table.change_foreign_keys()
            if table.name not in self.table_names:
                self.table_names.append(table.name)

    def get(self,table):
        if type(table) in [list,tuple]:
            return [self.get(x) for x in table]
        if isinstance(table,str):
            table_name = table.split('.')[-1]
        else:
            self.add_table(table)
            table_name=table.name
        return self.__dict__[table_name]

class DB:
    _engine=None
    _meta=None
    _sess=None
    _table_names=None
    def __init__(self,conn_str,schema=None,table_names=None):
        if os.path.isfile(conn_str):
            conn_str=f'sqlite:///{conn_str}'
        self.conn_str=conn_str
        self.schema=schema
        self._table_names=table_names
    def __str__(self):
        return f'{self.conn_str}'
    @property
    def table_names(self):
        if self._table_names is None:
            self._table_names=[x.split('.')[-1] for x in get_table_names(self.meta)]
        return self._table_names
    @property
    def engine(self,echo=False):
        if self._engine is None:
            self._engine=create_engine(self.conn_str,echo=echo)
        return self._engine
    @property
    def meta(self):
        if self._meta is None:
            self._meta=get_metadata(self.engine,schema=self.schema)
        return self._meta
    @property
    def sess(self):
        if self._sess is None:
            self._sess=get_session(self.engine)
        return self._sess
    def connect(self,verbose=1):
        if verbose>0:
            print(f'Connecting to {self.conn_str}...')
        if verbose>0:
            print('Connected!')
        return self
    def rollback(self):
        self.sess.rollback()
        return self
    @property
    def tables(self):
        if not hasattr(self,'_tables'):
            self._tables=tables(self)
        return self._tables
    def reset_tables(self):
        if hasattr(self,'_tables'):
            del self._tables



from .base import *
from .base_component import Component
from .dropdown import add_to_dropdown
import json, os, base64
from copy import copy

def parse_contents(content, filename, date,upload_to):
    data = content.encode("utf8").split(b";base64,")[1]
    full_path=os.path.join(upload_to,filename)
    with open(full_path, "wb") as fp:
        fp.write(base64.decodebytes(data))
    return filename,full_path

def upload_to_folder(contents, names, dates,upload_to, allowed_extensions=None):
    res={}
    if contents is not None:
        for c, n, d in zip(contents, names, dates):
            if allowed_extensions is None or any([n.lower().endswith(ext) for ext in allowed_extensions]):
                if upload_to is None:
                    new_name=new_path=n
                else:
                    new_name,new_path=parse_contents(c, n, d, upload_to)
                res[new_name]=new_path
    return res

class Uploader(Component):
    def __init__(self,type='area',multi=False,label=None,style=None,allowed_extensions=None,upload_to=None,dropdown=None,keep_old_uploads=True,condition_check=None,alert=None):
        if label is None:
            label='Files' if multi else 'File'

        if style is None: style={}
        if type=='area':
            children=html.Div(['Drag and Drop or ',html.A(f'Select {label}')])
            style.update({
                'width': '100%',
                'height': '60px',
                'lineHeight': '60px',
                'borderWidth': '1px',
                'borderStyle': 'dashed',
                'borderRadius': '5px',
                'textAlign': 'center',
                'margin-top':'10px'
            })
        elif type=='link':
            children=html.A(f'Select {label}')
        else:
            children=html.A(f'Upload {label}', className="btn btn-primary", style={'color':'#FFFFFF'})

        self.upload=dcc.Upload(children=children,style=style,multiple=multi)
        self.upload_info=html.Div('{}',style={'display':'none'})

        self.layout=html.Div([self.upload_info,self.upload])

        self.properties={'contents':self.upload.contents,
                        'filename':self.upload.filename,
                        'last_modified':self.upload.last_modified,
                        'info':self.upload_info.children}

        #callbacks
        @self([self.info]+([alert.alerts] if alert is not None else []),[self.contents,self.filename,self.last_modified],[self.info]+([alert.alerts] if alert is not None else []))
        def _upload_fcn(contents, names, dates, upload_data, alerts=None):
            if alerts is None: alerts=[]
            if not multi:
                contents=[contents]; names=[names]; dates=[dates]
            res=upload_to_folder(contents, names, dates,upload_to, allowed_extensions=allowed_extensions)

            msg='';color='primary'
            if condition_check is not None:
                msg,color,new_res=condition_check(copy(res))
                for path in list(res.values()):
                    if os.path.isfile(path) and path not in list(new_res.values()):
                        os.remove(path)
                    res=new_res

            if msg == '' and len(res)>0:
                color='success'
                msg=f'{", ".join(list(res))} {"were" if len(res)>1 else "was"} added successfully!'

            if keep_old_uploads:
                res.update(json.loads(upload_data))

            res=[json.dumps(res)]
            if alert is not None:
                alerts.append(alert.create(msg,color))
                res.append(alerts)

            return tuple(res)

        if dropdown is not None:
            self.properties.update({'value':dropdown.value,
                                    'options':dropdown.options})

            @self([dropdown.options,dropdown.value],self.info,[dropdown.options,dropdown.value])
            def upload_add_to_dropdown(upload_data,options,values):
                upload_data=json.loads(upload_data)
                new_options = [{'label': name, 'value': path} for name,path in upload_data.items()]
                new_values = list(upload_data.values())
                return add_to_dropdown(new_options,new_values,options,values,multi=multi,keep_old_uploads=keep_old_uploads)



import tkinter as tk
import threading

from pysts.utils.utils import create_logger, append_fcns


class Component(object):
    """ tkinter component
    """
    contexts = threading.local()
    def __init__(self):
        self.logger = create_logger(self.__class__.__name__)
        type(self).add_instance(self)
    @classmethod
    def get_logger(cls):
        if not hasattr(cls.contexts, '_logger'):
            cls.contexts._logger = create_logger('component')
        return cls.contexts._logger
    @classmethod
    def get_callbacks(cls):
        if not hasattr(cls.contexts, '_callbacks'):
            cls.contexts._callbacks = []
        return cls.contexts._callbacks
    @classmethod
    def _add_callback(cls,callback):
        callbacks=cls.get_callbacks()
        callbacks.append(callback)
    @classmethod
    def get_instances(cls):
        if not hasattr(cls.contexts, '_instances'):
            cls.contexts._instances=[]
        return cls.contexts._instances
    @classmethod
    def add_instance(cls,instance):
        instances=cls.get_instances()
        if instance not in instances:
            instances.append(instance)

    # Functions that may be used by instance
    @classmethod
    def callback(cls,component_func):
        def decorator(self,*event_args,**event_kwargs):
            no_args=False
            if len(event_args) == 1 and len(event_kwargs)==0 and callable(event_args[0]):
                callback_func=event_args[0]
                event_args=list(event_args)[1:]
                no_args=True

            def _decorator(callback_func):
                callback=component_func(self,callback_func,*event_args,**event_kwargs)
                if callback:
                    cls._add_callback(self,callback)
                return callback_func
            if no_args:
                # No arguments, this is the decorator (manually set defaults)
                return _decorator(callback_func)
            else:
                # This is just returning the decorator
                return _decorator
        return decorator

    # Functions that may be overloaded
    def stop(self):
        self.logger.info(f"Stopping jobs for {self.__class__.__name__}.")
        type(self)._stop_jobs(listeners=self)
    def start(self, block=False):
        self.logger.info(f"Starting jobs for {self.__class__.__name__}...")
        type(self)._start_jobs(block=block, listeners=self)



import sys
import threading, time
from datetime import timedelta
from .listener import Listener

class Job(threading.Thread):
    def __init__(self, listener, interval, execute, *args, **kwargs):
        threading.Thread.__init__(self)
        self.listener=listener
        self.daemon = False
        self.stopped = threading.Event()
        self.interval = interval
        self.execute = execute
        self.args = args
        self.kwargs = kwargs

    def stop(self):
        self.stopped.set()
        self.join()
    def start(self,block=False):
        self.daemon = not block
        super().start()
    def run(self):
        while not self.stopped.wait(self.interval.total_seconds()):
            res=self.execute(*self.args, **self.kwargs)
            if res is False:
                self.stop()

class Timeloop(Listener):
    @Listener.event
    def interval(self,func,seconds,*args,**kwargs):
        return Job(self,timedelta(seconds=seconds),execute=func,*args,**kwargs)

if __name__ == "__main__":
    tl = Timeloop()
    COUNTER=0
    @tl.interval(seconds=2)
    def sample_job_every_2s():
        global COUNTER
        COUNTER+=1
        print("{}: 2s job current time : {}".format(COUNTER, time.ctime()))
    tl.start(block=True)



from pysts.utils.modules import Lazy_Module

pynput=Lazy_Module('pynput')

class Keyboard(object):
    _controller=None
    @property
    def controller(self):
        if not self._controller:
            self._controller=pynput.keyboard.Controller()
        return self._controller

    @property
    def key(self):
        return pynput.keyboard.Key

    def press(self,key,release=False):
        self.controller.press(key)
        if release:
            self.release(key)

    def release(self,key):
        self.controller.release(key)

    def type(self,chars):
        self.controller.type(chars)

class Mouse(object):
    _controller=None
    @property
    def controller(self):
        if not self._controller:
            self._controller=pynput.mouse.Controller()
        return self._controller

    @property
    def position(self):
        return self.controller.position

    def click(self,left=True,button=None,double=True,num=None):
        num = num if num else (2 if double else 1)
        button=button if button else (Button.left if left else Button.right)
        self.controller.click(button, num)

    def press(self,left=True,button=None):
        button=button if button else (Button.left if left else Button.right)
        self.controller.press(button)

    def release(self,left=True,button=None):
        button=button if button else (Button.left if left else Button.right)
        self.controller.release(button)

    def move(self,x,y,relative=False):
        if relative:
            self.controller.move(x, y)
        else:
            self.controller.position = (x, y)

    def scroll(self,steps=1):
        self.controller.scroll(0, steps)



import threading, signal, time

from pysts.utils.utils import create_logger, append_fcns


class ServiceExit(Exception):
    """
    Custom exception which is used to trigger the clean exit
    of all running threads and the main program.
    """
    pass

def service_shutdown(signum, frame):
    raise ServiceExit

class Listener(object):
    """ Listerner class to run functions on events
        A listener may have multiple parallel jobs
        At a minimum, the listener should add one job if a callback is registered
        Each job may have multiple events to which callbacks may be allocated through the @self._event decorator
        E.g.,
            class Mouse(Listener):
                @Listener.event
                def on_click(self,x, y, button, pressed):
                    print(f"{'Pressed' if pressed else 'Released'} button {button}")
            mouse=Mouse()
            @mouse.on_click
            def my_func(x,y, button, pressed):
                print(f"it was at position ({x},{y})")

            class Timeloop(Listener):
                @Listener.event
                def interval(self,func,seconds,*args,**kwargs):
                    return Job(self,timedelta(seconds=seconds),execute=func,*args,**kwargs)

            tl=Timeloop()
            @tl.interval(seconds=2)
            def sample_job_every_2s():
                print("2s job current time : {}".format(COUNTER, time.ctime()))
    """
    contexts = threading.local()
    def __init__(self):
        self.logger = create_logger(self.__class__.__name__)
        type(self).add_instance(self)
    @classmethod
    def get_logger(cls):
        if not hasattr(cls.contexts, '_logger'):
            cls.contexts._logger = create_logger('listener')
        return cls.contexts._logger
    @classmethod
    def get_jobs(cls):
        if not hasattr(cls.contexts, '_jobs'):
            cls.contexts._jobs = {}
        return cls.contexts._jobs
    @classmethod
    def _add_job(cls,listener,job):
        clsjobs=cls.get_jobs()
        if listener not in clsjobs:
            clsjobs[listener]=[]
        if job not in clsjobs[listener]:
            clsjobs[listener].append(job)
    @classmethod
    def get_instances(cls):
        if not hasattr(cls.contexts, '_instances'):
            cls.contexts._instances=[]
        return cls.contexts._instances
    @classmethod
    def add_instance(cls,instance):
        instances=cls.get_instances()
        if instance not in instances:
            instances.append(instance)
    @classmethod
    def _block_main_thread(cls):
        signal.signal(signal.SIGTERM, service_shutdown)
        signal.signal(signal.SIGINT, service_shutdown)
        while cls.jobs_alive():
            try:
                time.sleep(1)
            except ServiceExit:
                cls._stop_jobs()
                break
    @classmethod
    def jobs_alive(cls):
        return len(cls.get_running_jobs())>0
    @classmethod
    def get_running_jobs(cls):
        return [job for listener,jobs in cls.get_jobs().items() for job in jobs if hasattr(job,'is_alive') and job.is_alive()]
    @classmethod
    def _get_listeners_jobs(cls,listeners=None):
        if listeners and type(listeners) not in [list,tuple]: listeners=[listeners]
        return [(listener,job) for listener,jobs in cls.get_jobs().items() if listeners is None or listener in listeners for job in jobs]
    @classmethod
    def _start_jobs(cls, block=False, listeners=None):
        for listener,j in cls._get_listeners_jobs(listeners):
            if callable(j):
                cls.get_jobs()[listener].remove(j)
                j=j()
                cls._add_job(listener,j)
            if 'block' in j.start.__code__.co_varnames:
                j.start(block)
            else:
                j.start()
            cls.get_logger().info("Started job {}".format(j))

        if block:
            cls._block_main_thread()

    @classmethod
    def _stop_jobs(cls, listeners=None):
        for listener,j in cls._get_listeners_jobs(listeners):
            cls.get_logger().info("Stopping job {}".format(j))
            j.stop()

    # Functions that may be used by instance
    @classmethod
    def event(cls,listener_func):
        def decorator(self,*event_args,**event_kwargs):
            no_args=False
            if len(event_args) == 1 and len(event_kwargs)==0 and callable(event_args[0]):
                callback_func=event_args[0]
                event_args=list(event_args)[1:]
                no_args=True

            def _decorator(callback_func):
                job=listener_func(self,callback_func,*event_args,**event_kwargs)
                if job:
                    cls._add_job(self,job)
                return callback_func
            if no_args:
                # No arguments, this is the decorator (manually set defaults)
                return _decorator(callback_func)
            else:
                # This is just returning the decorator
                return _decorator
        return decorator

    # Functions that may be overloaded
    def stop(self):
        self.logger.info(f"Stopping jobs for {self.__class__.__name__}.")
        type(self)._stop_jobs(listeners=self)
    def start(self, block=False):
        self.logger.info(f"Starting jobs for {self.__class__.__name__}...")
        type(self)._start_jobs(block=block, listeners=self)



import os
from .utils import *

def libre_convert_to_pdf(paths):
    LIBRE_OFFICE=''
    res=[]
    for path in paths:
        if path.endswith('docx'):
            pdf_path=path.replace('.docx','.pdf')
            p = Popen([LIBRE_OFFICE, '--headless', '--convert-to', 'pdf', '--outdir',
                       pdf_path, path])
            p.communicate()
            res.append(pdf_path)
    return res

def docx_to_pdf(paths):
    #docx2pdf=get_package('docx2pdf')
    pypandoc=get_package('pypandoc')
    res=[]
    for path in paths:
        if path.endswith('docx'):
            pdf_path=path.replace('.docx','.pdf')
            #docx2pdf.convert(path, pdf_path)
            pypandoc.convert_file(path, 'pdf', outputfile=pdf_path,extra_args=['--pdf-engine=xelatex'])
            res.append(pdf_path)
    return res

def create_word_from_html(html,template_path,output_path=None):
    pypandoc=get_package('pypandoc')
    docx=get_package('docx','python-docx')
    Document=docx.Document

    with open("temp.html",'w') as f:
        f.write(html)
    pypandoc.convert_file('temp.html', 'docx', outputfile="temp.docx")
    temp_doc=Document("temp.docx")

    if template_path is not None and os.path.isfile(template_path):
        document = Document(template_path)

        #Add elements
        for elem in temp_doc.element.body:
            document.element.body.append(elem)

        #Add image parts
        for imagepart in temp_doc.part._package.image_parts:
            document.part._package.image_parts.append(imagepart)
        for rId,rel in temp_doc.part.rels.items():
             if rel.reltype.endswith('image'):
                document.part.rels[rId]=rel
                document.part.rels._target_parts_by_rId[rId] = rel._target
    else:
        document=temp_doc

    #change table styles
    for table in document.tables:
        table.style = 'default'
        for row in table.rows:
            for cell in row.cells:
                for paragraph in cell.paragraphs:
                    paragraph.style='table_normal'

    if output_path is None:
        report_name=f'Report_{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}.docx'
        output_path=os.path.join(os.getcwd(),report_name)

    document.save(output_path)
    os.remove('temp.docx')
    os.remove('temp.html')

    return [output_path]



from setuptools import setup, find_packages
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as f:
    requirements = f.readlines()

setup(
  name = 'pysts',         # How you named your package folder (MyLib)
  packages=find_packages(exclude=['tests*']),
  version = '0.5.0',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Basic python utilities',   # Give a short description about your library
  author = 'Stephen Swart',                   # Type in your name
  url = 'https://github.com/sdswart/pysts.git',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/sdswart/pysts/archive/v_0.5.0.tar.gz',    # Used for pipy
  keywords = ['utilities',],   # Keywords that define your package best
  long_description=long_description,
  long_description_content_type="text/markdown",
  install_requires=requirements,
  classifiers=[
      "Programming Language :: Python :: 3",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
  ]
)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Purpose
    Demonstrate basic object operations in Amazon S3.
    Learn how to put, get, delete, and configure objects.
    Usage is shown in the test/test_object_wrapper.py file.

Running the tests
    The best way to learn how to use this service is to run the tests.
    For instructions on testing, see the README.

Running the code
    Run the usage_demo function in a command window or individual functions in
    the Python shell to make calls to your AWS account.
    For instructions on running the code, see the README.

Additional information
    Running this code might result in charges to your AWS account.
"""

import io
import json
import logging
import os
import random
import uuid

from botocore.exceptions import ClientError

from . import bucket_wrapper
from .utils import data_to_bytes

logger = logging.getLogger(__name__)

def create_temp_path(object_key,to_path=None):
    filename=os.path.basename(object_key)
    if to_path is None:
        to_path=filename
    num=0
    while os.path.exists(to_path):
        num+=1
        to_path=f'{filename}-{num}'
    return to_path

def copy_object_to_local(bucket,object_key,to_path=None):
    to_path=create_temp_path(object_key,to_path=to_path)

    body=get_object(bucket, object_key)

    with open(to_path,'wb') as f:
        f.write(body)
    return to_path

class s3_file(io.FileIO):
    def __init__(self,bucket,object_key,mode='r'):
        if object_exists(bucket, object_key):
            path=copy_object_to_local(bucket,object_key)
        else:
            path=create_temp_path(object_key)
            with open(path,'w') as f: pass
        self._bucket=bucket
        self._object_key=object_key
        super().__init__(path,mode)
    def close(self,*args,**kwargs):
        res=super().close(*args,**kwargs)
        if self.mode.startswith('r'): #save the file to s3
            with open(self.name,'rb') as f:
                put_object(self._bucket, self._object_key, f.read())
        os.remove(self.name)
        return res

def put_object(bucket, object_key, data):
    """
    Upload data to a bucket and identify it with the specified object key.

    Usage is shown in usage_demo at the end of this module.

    :param bucket: The bucket to receive the data.
    :param object_key: The key of the object in the bucket.
    :param data: The data to upload. This can either be bytes or a string. When this
                 argument is a string, it is interpreted as a file name, which is
                 opened in read bytes mode.
    """
    put_data = data_to_bytes(data)

    try:
        obj = bucket.Object(object_key)
        obj.put(Body=put_data)
        obj.wait_until_exists()
        logger.info("Put object '%s' to bucket '%s'.", object_key, bucket.name)
    except ClientError:
        logger.exception("Couldn't put object '%s' to bucket '%s'.",
                         object_key, bucket.name)
        raise
    finally:
        if getattr(put_data, 'close', None):
            put_data.close()


def get_object(bucket, object_key):
    """
    Gets an object from a bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket: The bucket that contains the object.
    :param object_key: The key of the object to retrieve.
    :return: The object data in bytes.
    """
    try:
        body = bucket.Object(object_key).get()['Body'].read()
        logger.info("Got object '%s' from bucket '%s'.", object_key, bucket.name)
    except ClientError:
        logger.exception(("Couldn't get object '%s' from bucket '%s'.",
                          object_key, bucket.name))
        raise
    else:
        return body


def open_object(bucket, object_key,mode='rb'):
    return s3_file(bucket, object_key,mode=mode)

def object_exists(bucket, object_key):
    """
    Gets an object from a bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket: The bucket that contains the object.
    :param object_key: The key of the object to retrieve.
    :return: The object data in bytes.
    """
    try:
        body = bucket.Object(object_key).load()
        return True
    except ClientError:
        logger.exception(("Couldn't load object '%s' from bucket '%s'.",
                          object_key, bucket.name))
        raise

    return False


def list_objects(bucket, prefix=None):
    """
    Lists the objects in a bucket, optionally filtered by a prefix.

    Usage is shown in usage_demo at the end of this module.

    :param bucket: The bucket to query.
    :param prefix: When specified, only objects that start with this prefix are listed.
    :return: The list of objects.
    """
    try:
        if not prefix:
            objects = list(bucket.objects.all())
        else:
            objects = list(bucket.objects.filter(Prefix=prefix))
        logger.info("Got objects %s from bucket '%s'",
                    [o.key for o in objects], bucket.name)
    except ClientError:
        logger.exception("Couldn't get objects for bucket '%s'.", bucket.name)
        raise
    else:
        return objects


def copy_object(source_bucket, source_object_key, dest_bucket, dest_object_key):
    """
    Copies an object from one bucket to another.

    Usage is shown in usage_demo at the end of this module.

    :param source_bucket: The bucket that contains the source object.
    :param source_object_key: The key of the source object.
    :param dest_bucket: The bucket that receives the copied object.
    :param dest_object_key: The key of the copied object.
    :return: The new copy of the object.
    """
    try:
        obj = dest_bucket.Object(dest_object_key)
        obj.copy_from(CopySource={
            'Bucket': source_bucket.name,
            'Key': source_object_key
        })
        obj.wait_until_exists()
        logger.info("Copied object from %s/%s to %s/%s.",
                    source_bucket.name, source_object_key,
                    dest_bucket.name, dest_object_key)
    except ClientError:
        logger.exception("Couldn't copy object from %s/%s to %s/%s.",
                         source_bucket.name, source_object_key,
                         dest_bucket.name, dest_object_key)
        raise
    else:
        return obj


def delete_object(bucket, object_key):
    """
    Removes an object from a bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket: The bucket that contains the object.
    :param object_key: The key of the object to delete.
    """
    try:
        obj = bucket.Object(object_key)
        obj.delete()
        obj.wait_until_not_exists()
        logger.info("Deleted object '%s' from bucket '%s'.", object_key, bucket.name)
        return True
    except ClientError:
        logger.exception("Couldn't delete object '%s' from bucket '%s'.",
                         object_key, bucket.name)
        raise


def delete_objects(bucket, object_keys):
    """
    Removes a list of objects from a bucket.
    This operation is done as a batch in a single request.

    Usage is shown in usage_demo at the end of this module.

    :param bucket: The bucket that contains the objects.
    :param object_keys: The list of keys that identify the objects to remove.
    :return: The response that contains data about which objects were deleted
             and any that could not be deleted.
    """
    try:
        response = bucket.delete_objects(Delete={
            'Objects': [{
                'Key': key
            } for key in object_keys]
        })
        if 'Deleted' in response:
            logger.info("Deleted objects '%s' from bucket '%s'.",
                        [del_obj['Key'] for del_obj in response['Deleted']],
                        bucket.name)
        if 'Errors' in response:
            logger.warning(
                "Sadly, could not delete objects '%s' from bucket '%s'.",
                [f"{del_obj['Key']}: {del_obj['Code']}"
                 for del_obj in response['Errors']],
                bucket.name)
    except ClientError:
        logger.exception()
        raise
    else:
        return response


def empty_bucket(bucket):
    """
    Remove all objects from a bucket.

    :param bucket: The bucket to empty.
    """
    try:
        bucket.objects.delete()
        logger.info("Emptied bucket '%s'.", bucket.name)
        return True
    except ClientError:
        logger.exception("Couldn't empty bucket '%s'.", bucket.name)
        raise


def put_acl(bucket, object_key, email):
    """
    Applies an ACL to an object that grants read access to an AWS user identified
    by email address.

    Usage is shown in usage_demo at the end of this module.

    :param bucket: The bucket that contains the object.
    :param object_key: The key of the object to update.
    :param email: The email address of the user to grant access.
    """
    try:
        acl = bucket.Object(object_key).Acl()
        # Putting an ACL overwrites the existing ACL, so append new grants
        # if you want to preserve existing grants.
        grants = acl.grants if acl.grants else []
        grants.append({
            'Grantee': {
                'Type': 'AmazonCustomerByEmail',
                'EmailAddress': email
            },
            'Permission': 'READ'
        })
        acl.put(
            AccessControlPolicy={
                'Grants': grants,
                'Owner': acl.owner
            }
        )
        logger.info("Granted read access to %s.", email)
    except ClientError:
        logger.exception("Couldn't add ACL to object '%s'.", object_key)
        raise


def get_acl(bucket, object_key):
    """
    Gets the ACL of an object.

    Usage is shown in usage_demo at the end of this module.

    :param bucket: The bucket that contains the object.
    :param object_key: The key of the object to retrieve.
    :return: The ACL of the object.
    """
    try:
        acl = bucket.Object(object_key).Acl()
        logger.info("Got ACL for object %s owned by %s.",
                    object_key, acl.owner['DisplayName'])
    except ClientError:
        logger.exception("Couldn't get ACL for object %s.", object_key)
        raise
    else:
        return acl


def usage_demo():
    """Demonstrated ways to use the functions in this module."""
    bucket = bucket_wrapper.create_bucket(
        'usage-demo-object-wrapper-' + str(uuid.uuid1()),
        bucket_wrapper.s3_resource.meta.client.meta.region_name)

    object_key = os.path.split(__file__)[-1]
    put_object(bucket, object_key, __file__)
    print(f"Put file object with key {object_key} in bucket {bucket.name}.")

    with open(__file__) as file:
        lines = file.readlines()

    for _ in range(10):
        line = random.randint(0, len(lines))
        put_object(bucket, f'line-{line}', bytes(lines[line], 'utf-8'))
    print(f"Put 10 random lines from this script as objects.")

    listed_lines = list_objects(bucket, 'line-')
    print(f"They are: {', '.join(l.key for l in listed_lines)}")

    line_to_delete = listed_lines.pop()
    line_body = get_object(bucket, line_to_delete.key)
    print(f"Got object with key {line_to_delete.key} and body {line_body}.")
    delete_object(bucket, line_to_delete.key)
    print(f"Deleted object with key {line_to_delete.key}.")

    copy_key = listed_lines[0].key + '-copy'
    copy_object(bucket, listed_lines[0].key, bucket, copy_key)
    print(f"Made a copy of object {listed_lines[0].key}, named {copy_key}.")

    try:
        put_acl(bucket, copy_key, 'arnav@example.net')
        acl = get_acl(bucket, copy_key)
        print(f"Put ACL grants on object {object_key}: {json.dumps(acl.grants)}")
    except ClientError as error:
        if error.response['Error']['Code'] == 'UnresolvableGrantByEmailAddress':
            print("Couldn't apply the ACL to the object because the specified "
                  "email is for a test user who does not exist. For this request to "
                  "succeed, you must replace the user email with one for an "
                  "actual AWS user.")
        else:
            raise

    empty_bucket(bucket)
    print(f"Emptied bucket {bucket.name} in preparation for deleting it.")

    bucket_wrapper.delete_bucket(bucket)
    print(f"Deleted bucket {bucket.name}.")


def main():
    go = input("Running the usage demonstration uses your default AWS account "
               "credentials and might incur charges on your account. Do you want "
               "to continue (y/n)? ")
    if go.lower() == 'y':
        print("Starting the usage demo. Enjoy!")
        usage_demo()
    else:
        print("Thanks anyway!")


if __name__ == '__main__':
    main()



from nbconvert.filters.markdown_mistune import IPythonRenderer, MarkdownWithMath
from nbconvert.preprocessors import ExecutePreprocessor
import nbformat, os, subprocess, json
from IPython import get_ipython
import matplotlib.pyplot as plt
from IPython.display import display,HTML
from IPython.utils.capture import RichOutput

from .utils import *

def get_notebook(nb):
    if isinstance(nb,str) and os.path.isfile(nb):
        with open(nb) as f:
            nb = nbformat.read(f, as_version=4)
    return nb

def get_notebook_inputs(nb):
    nb=get_notebook(nb)
    res={}
    for cell in nb['cells']:
        if cell['cell_type']=='raw':
            lines=cell['source'].split('\n')
            if lines[0].lower().startswith('inputs'):
                for line in lines:
                    if '=' in line:
                        parts=line.split('=')
                        res[parts[0]]=json.loads(parts[1])
    return res


def run_notebook(path,inputs=None,timeout=3600,destination=None,allow_errors=False):
    nb=get_notebook(path)

    if isinstance(inputs,dict):
        dict_to_environ(inputs)

    ep = ExecutePreprocessor(timeout=3600, kernel_name='python3')
    ep.allow_errors = allow_errors

    notebook_dir=os.path.split(path)[0]
    path_dir='./' if notebook_dir== '' else (os.path.relpath(os.path.split(path)[0],os.getcwd())+os.sep)
    ep.preprocess(nb, {'metadata': {'path': path_dir}})

    if destination is None:
        destination=path
    with open(destination, 'wt') as f:
        nbformat.write(nb, f)

def get_notebook_outputs(path,inputs=None,progress=None,info=None):
    notebook = nbformat.read(path, as_version=4) #metadata: "report":{"show_markdown_progress":true,"include_markdown":false}
    show_markdown_progress=notebook.metadata['report']['show_markdown_progress'] if 'report' in notebook.metadata and 'show_markdown_progress' in notebook.metadata['report'] else False
    include_markdown=notebook.metadata['report']['include_markdown'] if 'report' in notebook.metadata and 'include_markdown' in notebook.metadata['report'] else False
    len_cells=len(notebook['cells'])

    if isinstance(inputs,dict):
        dict_to_environ(inputs)

    get_ipython().run_line_magic('matplotlib', 'widget')
    plt.ioff()

    scope=None
    res=[]
    for i,cell in enumerate(notebook['cells']):
        if cell['cell_type']=='code':
            code='\n'.join([line for line in cell['source'].split('\n') if not line.startswith('%')])
            capture,scope=capture_code(code,scope=scope)
            outputs=capture.outputs
            get_outputs_figure_widgets(outputs);
            res.extend(outputs)
        elif cell['cell_type']=='markdown' and ((info is not None and show_markdown_progress) or include_markdown):
            attachments = cell.get('attachments', {})
            renderer = IPythonRenderer(escape=False, attachments=attachments,anchor_link_text='',exclude_anchor_links=True)
            html=MarkdownWithMath(renderer=renderer).render(cell['source'])
            if info is not None and show_markdown_progress:
                info.children.append(html)
            if include_markdown:
                html=HTML(html)
                res.append(RichOutput(data={'text/plain':str(html),'text/html':html.data}))

        if progress is not None:
            progress.value = float(i+1)/len_cells
    plt.ion()
    return res
